'''pytests for: eda

Using eda.main() entrypoint

If you want to run this, consider running from the root of opencos repo:
    > pytest --verbose opencos/*/*.py
    > python3 -m pytest --verbose opencos/*/*.py
    > python3 -m pytest -rx opencos/*/*.py
 which avoids using any pip installed opencos.eda
'''

# pylint: disable=R0801 # (similar lines in 2+ files)

import os
import shutil
import subprocess
from contextlib import redirect_stdout, redirect_stderr

import pytest

from opencos import eda, util, eda_tool_helper
from opencos.tests import helpers
from opencos.tests.helpers import eda_wrap, eda_sim_wrap, eda_elab_wrap, \
    Helpers


THISPATH = os.path.dirname(__file__)

def chdir_remove_work_dir(relpath):
    '''Changes dir to relpath, removes the work directories (eda.work, eda.export*)'''
    return helpers.chdir_remove_work_dir(THISPATH, relpath)

# Figure out what tools the system has available, without calling eda.main(..)
config, tools_loaded = eda_tool_helper.get_config_and_tools_loaded()

def can_run_eda_sim() -> bool:
    '''Returns True if we have any installed tool that can run: eda sim'''
    return helpers.can_run_eda_command('sim', config=config)

def can_run_eda_elab() -> bool:
    '''Returns True if we have any installed tool that can run: eda elab'''
    return helpers.can_run_eda_command('elab', config=config)

@pytest.mark.skipif(
    'verilator' not in tools_loaded and 'vivado' not in tools_loaded,
    reason="requires verilator OR vivado"
)
def test_args_sim_default_tool():
    '''Test that: eda sim <target>; works'''
    chdir_remove_work_dir('../../lib/tests')
    rc = eda_sim_wrap('oclib_fifo_test')
    print(f'{rc=}')
    assert rc == 0


class TestTargets(Helpers):
    '''Tests for: eda targets'''

    DEFAULT_DIR = os.path.join(THISPATH, '..', '..', 'lib', 'tests')

    def test_lib_tests__no_pattern(self):
        '''Test that this works: eda targets'''
        self.chdir()
        rc = self.log_it('targets --debug', use_eda_wrap=False)
        assert rc == 0
        assert self.is_in_log('oclib_fifo_test')
        assert self.is_in_log('oclib_rrarb_test')

    def test_lib_tests__with_pattern(self):
        '''Test that this works: eda targets oclib_fifo*test'''
        self.chdir()
        rc = self.log_it('targets oclib_fifo*test', use_eda_wrap=False)
        assert rc == 0
        assert self.is_in_log('oclib_fifo_test')
        assert not self.is_in_log('oclib_rrarb_test')


@pytest.mark.skipif(
    'verilator' not in tools_loaded, reason="requires verilator"
)
class TestsRequiresVerilator( # pylint: disable=too-many-public-methods
        Helpers
):
    '''Tests that require verilator, skip if not present (in some Github Action containers)'''

    def test_verilator_cant_run_synth(self):
        '''Checks eda.check_command_handler_cls(...) so we don't fallback to a different tool'''
        # If you say you want verilator, then we will NOT choose a different default handler.
        chdir_remove_work_dir('../../lib')
        rc = eda_wrap('synth', '--tool', 'verilator', 'oclib_fifo')
        print(f'{rc=}')
        assert rc != 0


    def test_args_sim(self):
        '''Basic sim with --tool verilator'''
        chdir_remove_work_dir('../../lib/tests')
        rc = eda_sim_wrap('--tool', 'verilator', 'oclib_fifo_test')
        print(f'{rc=}')
        assert rc == 0


    def test_args_sim_tool_with_path(self):
        '''Test for calling a tool as --tool=<tool>=</path/to/tool-exe>'''
        verilator_fullpath = shutil.which('verilator')
        chdir_remove_work_dir('../../lib/tests')
        rc = eda_sim_wrap('--tool', f'verilator={verilator_fullpath}', 'oclib_fifo_test')
        print(f'{rc=}')
        assert rc == 0

    def test_args_sim_with_coverage(self):
        '''Test for verilator --coverage'''
        chdir_remove_work_dir('../../lib/tests')
        rc = eda_wrap('sim', '--coverage', '--tool', 'verilator', 'oclib_fifo_test')
        print(f'{rc=}')
        assert rc == 0
        # We don't check the logs, but the command should succeed.

    def test_args_lint_only_sim(self):
        '''Confirm --lint-only works for Verilator with 'sim' command.'''
        chdir_remove_work_dir('../../lib/tests')
        rc = eda_sim_wrap('--lint-only', '--tool', 'verilator', 'oclib_fifo_test')
        print(f'{rc=}')
        assert rc == 0

    def test_args_elab(self):
        '''Test for: eda elab'''
        chdir_remove_work_dir('../../lib')
        rc = eda_elab_wrap('--tool', 'verilator', 'oclib_priarb')
        print(f'{rc=}')
        assert rc == 0

    def test_run_from_work_dir(self):
        '''
        Uses eda --stop-before-compile to craft the eda.work/(test)/ dirs and shell commands,
        and confirms that we can run those shell commands.
        '''

        chdir_remove_work_dir('../../lib/tests')
        rc = eda_sim_wrap('--stop-before-compile', '--tool', 'verilator', 'oclib_fifo_test')
        print(f'{rc=}')
        assert rc == 0

        os.chdir(os.path.join(THISPATH, '../../lib/tests/eda.work/oclib_fifo_test.sim'))
        res = subprocess.run(
            [ './lint_only.sh' ], stdout=subprocess.PIPE, stderr=subprocess.PIPE,
            check=True
        )
        rc = res.returncode
        print(f'{rc=}')
        assert rc == 0
        assert res.stdout
        assert res.stderr == b''

        res = subprocess.run(
            [ './all.sh' ], stdout=subprocess.PIPE, stderr=subprocess.PIPE,
            check=True
        )
        rc = res.returncode
        print(f'{rc=}')
        assert rc == 0
        assert res.stdout
        assert res.stderr == b''

        res = subprocess.run(
            [ './simulate_only.sh' ], stdout=subprocess.PIPE, stderr=subprocess.PIPE,
            check=True
        )
        rc = res.returncode
        print(f'{rc=}')
        assert rc == 0
        assert res.stdout
        assert res.stderr == b''

    def test_args_sim_waves(self):
        '''Test that --waves for verilator works (FST, not VCD)'''
        chdir_remove_work_dir('../../lib/tests')
        rc = eda_sim_wrap('--tool', 'verilator', '--waves', 'oclib_fifo_test')
        print(f'{rc=}')
        assert os.path.exists(os.path.join('.', 'eda.work', 'oclib_fifo_test.sim', 'dump.fst'))
        assert rc == 0

    def test_args_sim_waves_dumpvcd(self):
        '''Test that --waves --dump-vcd works for the opencos/ style SV tests'''
        chdir_remove_work_dir('../../lib/tests')
        rc = eda_sim_wrap('--tool', 'verilator', '--waves', '--dump-vcd', 'oclib_fifo_test')
        print(f'{rc=}')
        assert os.path.exists(os.path.join('.', 'eda.work', 'oclib_fifo_test.sim', 'dump.vcd'))
        assert rc == 0

    def test_args_sim_dumpvcd_verilator_trace(self):
        '''Do not set --dump-vcd, set --waves and do directly set +trace=vcd,
        and confirm +trace works as a bare CLI plusarg'''

        chdir_remove_work_dir('../../lib/tests')
        rc = self.log_it('sim --tool verilator --waves +trace=vcd oclib_fifo_test')
        assert rc == 0
        assert os.path.exists(os.path.join('.', 'eda.work', 'oclib_fifo_test.sim', 'dump.vcd'))
        lines = self.get_log_lines_with('exec: ./obj_dir/sim.exe')
        assert len(lines) == 1
        assert ' +trace=vcd ' in lines[0]
        assert ' +trace ' not in lines[0]

    def test_args_seed1(self):
        '''Test for: eda sim --tool verilator --seed <value>'''
        chdir_remove_work_dir('../../lib/tests')
        rc = self.log_it('sim --tool verilator --seed=1 oclib_fifo_test')
        assert rc == 0
        lines = self.get_log_lines_with('exec: ./obj_dir/sim.exe')
        assert len(lines) == 1
        assert ' +verilator+seed+1 ' in lines[0]

    def test_args_wno_fatal(self):
        '''Test for: eda sim --tool verilator w/ --verilate-args'''
        chdir_remove_work_dir('../../lib/tests')
        rc = self.log_it('sim --tool verilator --verilate-args=-Wno-fatal oclib_fifo_test')
        assert rc == 0
        lines = self.get_log_lines_with('exec: ')
        assert len(lines) == 2
        assert 'verilator' in lines[0]
        assert ' -Wno-fatal ' in lines[0]
        assert 'sim.exe' in lines[1]

    def test_args_sim_should_fail(self):
        '''Test that our command handler for verilator will fail b/c --xilinx set.'''
        chdir_remove_work_dir('../../lib/tests')
        # We'd expect this to fail b/c --xilinx and --tool verilator flags an error. I do not
        # want to use the xfail pytest decorator.
        rc = eda_sim_wrap('--xilinx', '--tool', 'verilator', 'oclib_fifo_test')
        print(f'{rc=}')
        assert rc != 0

    def test_more_plusargs_sim(self):
        '''Test that unprocessed plusargs become sim-plusargs on CLI'''
        chdir_remove_work_dir('../../lib/tests')
        rc = self.log_it('sim --tool verilator +info=300 +some_plusarg_novalue oclib_fifo_test')
        assert rc == 0
        lines = self.get_log_lines_with('exec: ./obj_dir/sim.exe')
        assert len(lines) == 1
        assert ' +info=300 ' in lines[0]
        assert ' +info ' not in lines[0]
        assert ' +some_plusarg_novalue ' in lines[0]
        assert ' +some_plusarg_novalue=' not in lines[0]

    def test_args_multi_sim(self):
        '''Basic test for: eda multi sim, with common args, should pass.'''
        chdir_remove_work_dir('../../lib/tests')
        rc = eda.main('multi', 'sim', '--fail-if-no-targets', '--seed=1', '--tool', 'verilator',
                      'oclib_fifo_test')
        print(f'{rc=}')
        assert rc == 0

    def test_args_multi_sim_timeout(self):
        '''Test for --single-timout in: eda multi'''
        chdir_remove_work_dir('../../lib/tests')
        rc = eda.main('multi', 'sim', '--fail-if-no-targets', '--seed=1', '--tool', 'verilator',
                      '--single-timeout', '10', 'oclib_fifo_test')
        print(f'{rc=}')
        assert rc == 0

    def test_args_multi_sim_should_fail(self):
        '''Checks that: eda multi --fail-if-no-targets; will fail b/c the found target fails'''
        chdir_remove_work_dir('../../lib/tests')
        # We'd expect this to fail b/c --xilinx and --tool verilator flags an error. I do not
        # want to use the xfail pytest decorator.
        rc = eda.main('multi', 'sim', '--fail-if-no-targets', '--seed=1', '--xilinx', '--tool',
                      'verilator', 'oclib_fifo_test')
        print(f'{rc=}')
        assert rc != 0

    def test_args_multi_sim_no_targets_should_fail(self):
        '''Checks that: eda multi --fail-if-no-targets; will fail if no targets expanded'''
        chdir_remove_work_dir('../../lib/tests')
        # We'd expect this to fail b/c no_targets* should expand to nothing.
        rc = eda.main('multi', 'sim', '--fail-if-no-targets', '--seed=1', '--tool', 'verilator',
                      'no_targets*')
        print(f'{rc=}')
        assert rc != 0

    def test_elab_verilator_no_deps_files_involved(self):
        '''Test that inferring the --top from last file in provides files works.'''
        # no --top set, have to infer its final file name.
        chdir_remove_work_dir('../../lib')

        cmd_list = (
            'elab --tool verilator +incdir+.. oclib_assert_pkg.sv oclib_pkg.sv'
            ' ../sim/ocsim_pkg.sv ../sim/ocsim_urand.sv ./rams/oclib_ram1r1w_infer.sv'
            ' ./rams/oclib_ram1r1w_infer_core.v oclib_fifo.sv'
        ).split()

        rc = eda.main(*cmd_list)
        print(f'{rc=}')
        assert rc == 0
        # We don't get a log for this, but we can check the output generated eda_output_config.yml.
        eda_config_yml_path = os.path.join(
            os.getcwd(), 'eda.work', 'oclib_fifo.elab', 'eda_output_config.yml'
        )
        data = util.yaml_safe_load(eda_config_yml_path)
        assert 'args' in data
        assert data['args'].get('top', '') == 'oclib_fifo'
        assert 'config' in data
        assert 'eda_original_args' in data['config']
        assert 'oclib_fifo.sv' in data['config']['eda_original_args']
        assert data.get('target', '') == 'oclib_fifo'

    def test_elab_verilator_some_deps_files_involved(self):
        '''Test calling targets (not files) on CLI.'''
        # no --top set, have to infer its final file name.
        chdir_remove_work_dir('../../lib')
        cmd_list = 'elab --tool verilator +incdir+.. all_pkg oclib_ram1r1w oclib_fifo.sv'.split()
        rc = eda.main(*cmd_list)
        print(f'{rc=}')
        assert rc == 0
        # We don't get a log for this, but we can check the output generated eda_output_config.yml.
        eda_config_yml_path = os.path.join(
            os.getcwd(), 'eda.work', 'oclib_fifo.elab', 'eda_output_config.yml'
        )
        data = util.yaml_safe_load(eda_config_yml_path)
        assert 'args' in data
        assert data['args'].get('top', '') == 'oclib_fifo'
        assert 'config' in data
        assert 'eda_original_args' in data['config']
        assert 'oclib_fifo.sv' in data['config']['eda_original_args']
        assert 'all_pkg' in data['config']['eda_original_args']
        assert 'oclib_ram1r1w' in data['config']['eda_original_args']
        assert data.get('target', '') == 'oclib_fifo'

    def test_elab_verilator_no_deps_files_involved_should_fail(self):
        '''Test using no DEPS file on file that doesn't exist.'''
        chdir_remove_work_dir('../../lib')
        # pick some non-existent file oclib_doesnt_exist.nope.sv
        cmd_list = 'elab --tool verilator +incdir+.. oclib_doesnt_exist.nope.sv'.split()
        cmd_list +=' oclib_assert_pkg.sv oclib_pkg.sv oclib_fifo.sv'.split()
        rc = eda.main(*cmd_list)
        print(f'{rc=}')
        assert rc != 0

    def test_config_reduced_yml(self):
        '''Test using provided EDA --config-yml=eda_config_reduced.yml, confirm installed w/ pip'''
        chdir_remove_work_dir('../../lib/tests')
        rc = eda_sim_wrap('--config-yml', 'eda_config_reduced.yml', '--tool', 'verilator',
                          'oclib_fifo_test')
        print(f'{rc=}')
        assert rc == 0

    def test_config_max_verilator_waivers_yml(self):
        '''Test using provided EDA --config-yml, confirm installed w/ pip'''
        chdir_remove_work_dir('../../lib/tests')
        rc = eda_sim_wrap('--config-yml', 'eda_config_max_verilator_waivers.yml', '--tool',
                          'verilator', 'oclib_fifo_test')
        print(f'{rc=}')
        assert rc == 0

    def test_config_yml_custom(self):
        '''Test using user-custom --config-yml arg'''
        chdir_remove_work_dir('../../lib/tests')
        rc = eda_sim_wrap('--config-yml', '../../opencos/tests/custom_config.yml', '--tool',
                          'verilator', 'oclib_fifo_test')
        print(f'{rc=}')
        assert rc == 0
        eda_config_yml_path = os.path.join(os.getcwd(), 'eda.work', 'oclib_fifo_test.sim',
                                           'eda_output_config.yml')
        data = util.yaml_safe_load(eda_config_yml_path)
        # make sure this config was actually used. We no longer re-add it to args
        # (it won't show up in 'original_args') it will will show up in the config though:
        used_yml_fname = data['config']['config-yml']
        assert used_yml_fname.endswith('opencos/tests/custom_config.yml')
        # this config overrides a value to False:
        assert 'config' in data
        local_config = data['config']
        assert local_config['dep_command_enables']['shell'] is False

    def test_verilator_rtl_missing_dumpfile_fst(self):
        '''test for eda with --waves, on RTL that has no $dumpfile,
        and we should auto-add _waves_pkg.sv with dump.fst'''

        chdir_remove_work_dir('./deps_files/test_deps_noext')
        rc = eda_wrap('sim', '--tool', 'verilator', '--waves', 'target_test')
        print(f'{rc=}')
        assert rc == 0

        assert not os.path.exists(
            os.path.join(os.getcwd(), 'eda.work', 'target_test.sim', 'dump.vcd')
        )
        assert os.path.exists(
            os.path.join(os.getcwd(), 'eda.work', 'target_test.sim', 'dump.fst')
        )

    def test_verilator_rtl_missing_dumpfile_vcd(self):
        '''test for eda with --waves, on RTL that has no $dumpfile,
        and we should auto-add _waves_pkg.sv with dump.vcd'''

        chdir_remove_work_dir('./deps_files/test_deps_noext')
        rc = eda_wrap('sim', '--tool', 'verilator', '--waves', '+trace=vcd', 'target_test')
        print(f'{rc=}')
        assert rc == 0

        assert os.path.exists(
            os.path.join(os.getcwd(), 'eda.work', 'target_test.sim', 'dump.vcd')
        )
        assert not os.path.exists(
            os.path.join(os.getcwd(), 'eda.work', 'target_test.sim', 'dump.fst')
        )

    def test_verilator_rtl_missing_dumpfile_none(self):
        '''test for eda with NO --waves, on RTL that has no $dumpfile,
        and we should see no auto-added _waves_pkg.sv and no dump.[vcd|fst]'''

        chdir_remove_work_dir('./deps_files/test_deps_noext')
        rc = eda_wrap('sim', '--tool', 'verilator', 'target_test')
        print(f'{rc=}')
        assert rc == 0

        assert not os.path.exists(
            os.path.join(os.getcwd(), 'eda.work', 'target_test.sim', 'dump.vcd')
        )
        assert not os.path.exists(
            os.path.join(os.getcwd(), 'eda.work', 'target_test.sim', 'dump.fst')
        )


class TestMissingDepsFileErrorMessages(Helpers):
    '''Test for missing DEPS.yml file, using 'eda export' to avoid tools.'''
    DEFAULT_DIR = os.path.join(THISPATH, 'deps_files', 'no_deps_here', 'empty')
    DEFAULT_LOG = 'eda.log'

    def test_bad0(self):
        '''Looks for target_bad0, but there is no DEPS file in .'''
        self.chdir()
        rc = self.log_it(command_str='export target_bad0')
        assert rc != 0
        assert self.is_in_log(
            'Trying to resolve command-line target=./target_bad0: was not found',
            'in deps_file=None, possible targets in deps file = []',
            windows_path_support=True
        )


class TestDepsResolveErrorMessages(Helpers):
    '''Tests to check that error messaging with DEPS.yml works as expected.

    If this was not checked, we could have failing tests that produce a stacktrace
    or other less helpful information to the user. This confirms that file/target/
    linenumber information is printed when available.'''

    DEFAULT_DIR = os.path.join(THISPATH, 'deps_files', 'error_msgs')
    DEFAULT_LOG = 'eda.log'

    # files foo.sv, foo2.sv, target_bad0.sv, and target_bad1.sv exist.
    # files missing*.sv and targets missing* do not exist.
    # These all "export" targets, to avoid requiring an installed tool (for example, to elab)

    def test_good0(self):
        '''Simple test with good target (foo)'''
        self.chdir()
        rc = self.log_it('export foo')
        assert rc == 0

    def test_good1(self):
        '''Simple test with good target (foo2)'''
        self.chdir()
        rc = self.log_it('export foo2')
        assert rc == 0

    def test_good2(self):
        '''Simple test with good target (foo + top=foo using deps str)'''
        self.chdir()
        rc = self.log_it('export target_good2')
        assert rc == 0

    def test_good3(self):
        '''Simple test with good target (foo2 + top=foo2 using deps list)'''
        self.chdir()
        rc = self.log_it('export target_good3')
        assert rc == 0

    # Bit of a change-detector-test here, but I want to make sure the
    # line= numbers get reported correctly for the calling target.
    def test_bad0(self):
        '''Tests missing file in DEPS target using implicit deps str style'''
        self.chdir()
        rc = self.log_it(command_str='export target_bad0')
        assert rc != 0
        assert self.is_in_log(
            "target=./missing0.sv (file?): called from ./DEPS.yml::target_bad0::line=20,",
            "File=missing0.sv not found in directory=.",
            windows_path_support=True
        )

    def test_bad1(self):
        '''Tests missing file in DEPS target using implicit deps list style'''
        self.chdir()
        rc = self.log_it(command_str='export target_bad1')
        assert rc != 0
        assert self.is_in_log(
            "target=./missing1.sv (file?): called from ./DEPS.yml::target_bad1::line=24,",
            "File=missing1.sv not found in directory=.",
            windows_path_support=True
        )

    def test_bad2(self):
        '''Tests missing file in DEPS target using deps as str style'''
        self.chdir()
        rc = self.log_it(command_str='export target_bad2')
        assert rc != 0
        assert self.is_in_log(
            "target=./missing2.sv (file?): called from ./DEPS.yml::target_bad2::line=28,",
            "File=missing2.sv not found in directory=.",
            windows_path_support=True
        )

    def test_bad3(self):
        '''Tests missing file in DEPS target using deps as list style'''
        self.chdir()
        rc = self.log_it(command_str='export target_bad3')
        assert rc != 0
        assert self.is_in_log(
            "target=./missing3.sv (file?): called from ./DEPS.yml::target_bad3::line=33,",
            "File=missing3.sv not found in directory=.",
            windows_path_support=True
        )

    def test_bad4(self):
        '''EDA on a bad target (bad target within deps of 'target_bad4'), explicit deps str'''
        self.chdir()
        rc = self.log_it(command_str='export target_bad4')
        assert rc != 0
        assert self.is_in_log(
            "target=./missing_target4: called from ./DEPS.yml::target_bad4::line=39,",
            "Target not found in deps_file=./DEPS.yml",
            windows_path_support=True
        )

    def test_bad5(self):
        '''EDA on a bad target (bad target within deps of 'target_bad4'), explicit deps list'''
        self.chdir()
        rc = self.log_it(command_str='export target_bad5')
        assert rc != 0
        assert self.is_in_log(
            "target=./missing_target5: called from ./DEPS.yml::target_bad5::line=43,",
            "Target not found in deps_file=./DEPS.yml",
            windows_path_support=True
        )

    def test_bad6(self):
        '''EDA on a bad target (bad target within deps of 'target_bad4'), deps str'''
        self.chdir()
        rc = self.log_it(command_str='export target_bad6')
        assert rc != 0
        assert self.is_in_log(
            "target=./missing_target6: called from ./DEPS.yml::target_bad6::line=47,",
            "Target not found in deps_file=./DEPS.yml",
            windows_path_support=True

        )

    def test_bad7(self):
        '''EDA on a bad target (bad target within deps of 'target_bad4'), deps list'''
        self.chdir()
        rc = self.log_it(command_str='export target_bad7')
        assert rc != 0
        assert self.is_in_log(
            "target=./missing_target7: called from ./DEPS.yml::target_bad7::line=52,",
            "Target not found in deps_file=./DEPS.yml",
            windows_path_support=True
        )

    def test_cmd_line_good0(self):
        '''EDA w/ out DEPS, on file'''
        self.chdir()
        rc = self.log_it(command_str='export foo.sv')
        assert rc == 0

    def test_cmd_line_good1(self):
        '''EDA w/ out DEPS, on two files'''
        self.chdir()
        rc = self.log_it(command_str='export foo.sv foo2.sv')
        assert rc == 0

    def test_cmd_line_bad0(self):
        '''EDA calling a non-existent target in DEPS file'''
        self.chdir()
        rc = self.log_it(command_str='export nope_target0')
        assert rc != 0
        assert self.is_in_log(
            "Trying to resolve command-line target=./nope_target0: was not",
            "found in deps_file=./DEPS.yml, possible targets in deps file = ['foo'",
            windows_path_support=True
        )

    def test_cmd_line_bad1(self):
        '''EDA calling a non-existent target in DEPS file, with file that exists.'''
        self.chdir()
        rc = self.log_it(command_str='export foo.sv nope_target1')
        assert rc != 0
        assert self.is_in_log(
            "Trying to resolve command-line target=./nope_target1: was not",
            "found in deps_file=./DEPS.yml, possible targets in deps file = ['foo'",
            windows_path_support=True
        )

    def test_cmd_line_bad2(self):
        '''EDA calling a non-existent file w/out DEPS'''
        self.chdir()
        rc = self.log_it(command_str='export nope_file0.sv')
        assert rc != 0
        assert self.is_in_log(
            "Trying to resolve command-line target=./nope_file0.sv",
            "(file?): File=nope_file0.sv not found in directory=.",
            windows_path_support=True
        )

    def test_cmd_line_bad3(self):
        '''EDA calling a non-existent file w/out DEPS, and a file that does exist.'''
        self.chdir()
        rc = self.log_it(command_str='export foo2.sv nope_file1.sv')
        assert rc != 0
        assert self.is_in_log(
            "Trying to resolve command-line target=./nope_file1.sv",
            "(file?): File=nope_file1.sv not found in directory=.",
            windows_path_support=True
        )


@pytest.mark.skipif('iverilog' not in tools_loaded, reason="requires iverilog")
class TestsRequiresIVerilog(Helpers):
    '''Test for Icarus Verilog'''

    def test_iverilog_help(self):
        '''Test for help'''
        rc = self.log_it('sim --tool iverilog help', use_eda_wrap=False)
        print(f'{rc=}')
        assert rc == 0
        assert self.is_in_log('Detected iverilog')
        assert self.is_in_log("Generic help for command='sim' (using 'CommandSimIverilog')")

    def test_iverilog_sim(self):
        '''Test for command sim'''
        chdir_remove_work_dir('deps_files/iverilog_test')
        cmd_list = 'sim --tool iverilog target_test'.split()
        rc = eda.main(*cmd_list)
        print(f'{rc=}')
        assert rc == 0


@pytest.mark.skipif(not can_run_eda_sim(), reason='no tool found to handle command: sim')
class TestDepsReqs:
    '''Tests for 'reqs' in the DEPS files. 'reqs' are requirements, like a .pcap or file
    used by SV $readmemh. They do not fit into normal compilable files (.sv, .v, .vhd[l])
    but are needed for export (command) and --export* (args on non-export command).'''

    def test_deps_reqs(self):
        '''Basic test using a .mem file in the 'deps' section of a target '''
        chdir_remove_work_dir('deps_files/non_sv_reqs')
        cmd_list = 'sim foo_test'.split()
        rc = eda.main(*cmd_list)
        assert rc == 0

    def test_deps_reqs2(self):
        '''Basic test using a .mem file in the 'reqs' section of a target'''
        chdir_remove_work_dir('deps_files/non_sv_reqs')
        cmd_list = 'sim foo_test2'.split()
        rc = eda.main(*cmd_list)
        assert rc == 0

    def test_deps_reqs3(self):
        '''Basic test using a .svh file in the 'reqs' section of a target'''
        chdir_remove_work_dir('deps_files/non_sv_reqs')
        cmd_list = 'sim foo_test3'.split()
        rc = eda.main(*cmd_list)
        assert rc == 0

    def test_deps_reqs4(self):
        '''Basic test using a .svh file with just incdirs (no reqs).'''
        chdir_remove_work_dir('deps_files/non_sv_reqs')
        cmd_list = 'sim foo_test4'.split()
        rc = eda.main(*cmd_list)
        assert rc == 0

    def test_deps_reqs5(self):
        '''Test that should fail due to reqs (bad file in reqs section)'''
        chdir_remove_work_dir('deps_files/non_sv_reqs')
        cmd_list = 'sim should_fail_foo_test5'.split()
        rc = eda.main(*cmd_list)
        assert rc != 0

    def test_deps_reqs6(self):
        '''Test that should fail due bad file in deps section (none in reqs)'''
        chdir_remove_work_dir('deps_files/non_sv_reqs')
        cmd_list = 'sim should_fail_foo_test6'.split()
        rc = eda.main(*cmd_list)
        assert rc != 0


@pytest.mark.skipif(not can_run_eda_sim(), reason='no tool found to handle command: sim')
def test_deps_command_order():
    '''Test for various "commands" within a DEPS target. This test checks that command
    order is preserved in the top-to-bottom deps order, meaning that eda.py has to collect
    all commands deps order, and then execute them in that order.'''

    chdir_remove_work_dir('deps_files/command_order')
    cmd_list = 'sim --stop-before-compile target_test'.split()
    with open('eda.log', 'w', encoding='utf-8') as f:
        with redirect_stdout(f):
            with redirect_stderr(f):
                rc = eda.main(*cmd_list)

    print(f'{rc=}')
    assert rc == 0

    # We should see "hi" before "bye" to confirm deps + command order is correct.
    # see ./deps_files/command_order/DEPS.yml - target = target_test
    found_str_list = [
        'exec: echo "hi"',
        'exec: echo "bye"',
    ]
    found_lines_list = [None, None]

    with open('eda.log', encoding='utf-8') as f:
        for lineno, line in enumerate(f.readlines()):
            line = line.rstrip()
            for idx,key in enumerate(found_str_list):
                if key in line:
                    found_lines_list[idx] = lineno

    assert found_lines_list[0] # found hi
    assert found_lines_list[1] # found bye
    assert found_lines_list[0] < found_lines_list[1] # hi before bye

    # Added check, we redirected to create eda.log earlier to confirm the targets worked,
    # but as a general eda.py check, all shell commands should create their own
    # {target}__shell_0.log file:
    work_dir = os.path.join(THISPATH, 'deps_files/command_order', 'eda.work', 'target_test.sim')
    with open(os.path.join(work_dir, 'target_echo_hi__shell_0.log'), encoding='utf-8') as f:
        text = ''.join(f.readlines()).strip()
        assert text == 'hi'
    # Added check, one of the targets uses a custom 'tee' file name, instead of the default log.
    with open(os.path.join(work_dir, 'custom_tee_echo_bye.log'), encoding='utf-8') as f:
        text = ''.join(f.readlines()).strip()
        assert text == 'bye'


@pytest.mark.skipif('verilator' not in tools_loaded, reason="requires verilator")
class TestDepsOtherMarkup:
    '''Tests for DEPS files that aren't YAML file extension.'''

    def test_deps_toml(self):
        '''test for DEPS.toml'''
        chdir_remove_work_dir('./deps_files/test_deps_toml')
        rc = eda_wrap('sim', '--tool', 'verilator', 'target_test')
        print(f'{rc=}')
        assert rc == 0

    def test_deps_json(self):
        '''test for DEPS.json'''
        chdir_remove_work_dir('./deps_files/test_deps_json')
        rc = eda_wrap('sim', '--tool', 'verilator', 'target_test')
        print(f'{rc=}')
        assert rc == 0

    def test_deps_no_extension(self):
        '''test for DEPS, which is treated as YAML'''
        chdir_remove_work_dir('./deps_files/test_deps_noext')
        rc = eda_wrap('sim', '--tool', 'verilator', 'target_test')
        print(f'{rc=}')
        assert rc == 0

@pytest.mark.skipif('verilator' not in tools_loaded, reason="requires verilator")
class TestForceFileExt(Helpers):
    '''Tests for treating non .sv files (.txt) as SV using sv@<file>'''

    def test_sv_at(self):
        '''test for file as sv@<file>'''
        chdir_remove_work_dir('./deps_files/force_file_ext')
        rc = self.log_it('sim --tool verilator sv@foo.txt')
        assert rc == 0
        assert self.is_in_log("force_file_ext/foo.txt:6: Verilog $finish")

    def test_v_at(self):
        '''test for file as v@<file>'''
        chdir_remove_work_dir('./deps_files/force_file_ext')
        rc = self.log_it('sim --tool verilator v@foo.txt')
        assert rc == 0
        assert self.is_in_log("force_file_ext/foo.txt:6: Verilog $finish")


@pytest.mark.skipif('verilator' not in tools_loaded, reason="requires verilator")
class TestDepsNoFilesTargets(Helpers):
    '''series of tests for running EDA w/out a DEPS target, all CLI files.'''

    def test_eda_sim__use_implicit_one_target(self):
        '''This test should work if the DEPS markup has a single target only'''
        # Using this b/c DEPS.toml has single target.
        chdir_remove_work_dir('./deps_files/test_deps_toml')
        rc = self.log_it('sim --tool verilator')
        assert rc == 0
        # Confirm the 'target_test' was used.
        assert self.is_in_log("using 'target_test' from")
        exec_lines = self.get_log_lines_with('exec: ')
        assert 'verilator ' in exec_lines[0]
        assert self.is_in_log("test_deps_toml/foo.sv:6: Verilog $finish")

    def test_eda_sim__wrong_target_shouldfail(self):
        '''This test should fail, wrong target name'''
        # Using this b/c DEPS.toml has single target.
        chdir_remove_work_dir('./deps_files/test_deps_toml')
        rc = eda_wrap('sim', '--tool', 'verilator', 'target_whoops')
        assert rc == 1

    def test_eda_sim__no_files_or_targets_shouldfail(self):
        '''This test should fail, there is DEPS.yml (empty, no implicit target), or missing file'''
        chdir_remove_work_dir('./deps_files/no_deps_here')
        rc = eda_wrap('sim', '--tool', 'verilator')
        assert rc == 1

    def test_eda_sim__no_files_or_targets_with_top_shouldfail(self):
        '''This test should fail, there is DEPS.yml (empty, no implicit target), or missing file'''
        chdir_remove_work_dir('./deps_files/no_deps_here')
        rc = eda_wrap('sim', '--tool', 'verilator', '--top', 'empty_file')
        assert rc == 1


class TestDepsTags(Helpers):
    '''Series of tests for DEPS - target - tags, in ./deps_files/tags_with_tools'''
    DEFAULT_DIR = os.path.join(THISPATH, 'deps_files', 'tags_with_tools')
    DEFAULT_LOG = 'eda.log'

    @pytest.mark.skipif('verilator' not in tools_loaded, reason="requires verilator")
    def test_tags_with_tools_verilator(self):
        '''test for DEPS target that hits with-tools: verilator, so that
        additional args are applied from the DEPS tag.'''
        self.chdir()
        logfile = 'verilator_eda.log'
        rc = self.log_it('sim --tool verilator target_test', logfile=logfile)
        assert rc == 0

        # so the full sim should have not run
        exec_lines = self.get_log_lines_with('exec: ', logfile=logfile)
        assert len(exec_lines) == 1, \
            f'{exec_lines=} should have only been the compile --lint-only'
        assert 'exec: ' in exec_lines[0] and 'verilator ' in exec_lines[0], \
            f'{exec_lines[0]=} should have been verilator compile'


    @pytest.mark.skipif('verilator' not in tools_loaded, reason="requires verilator")
    def test_tags_with_tools_replace_config_tools_verilator(self):
        '''test for DEPS target with tags that perfoms replacement config.

        AKA, lets you replace all the Verilator waivers to the DEPS target that only affect
        with-tools: verilator'''
        self.chdir()
        logfile = 'target_with_replace_config_tools_test.log'
        rc = self.log_it('sim --tool verilator target_with_replace_config_tools_test',
                         logfile=logfile)
        assert rc == 0
        # This target overrode all the Verilator waivers to nothing, so
        # we should see zero -Wno- in the log for verilator exec lines.
        exec_lines = self.get_log_lines_with('exec: ', logfile=logfile)
        assert len(exec_lines) == 2
        assert not '-Wno-' in exec_lines[0], f'-Wno- expected to be in one of: {exec_lines=}'


    @pytest.mark.skipif('verilator' not in tools_loaded, reason="requires verilator")
    def test_tags_with_tools_additive_config_tools_verilator(self):
        '''test for DEPS target with tags that perfoms additive config.

        AKA, lets you add Verilator waivers to the DEPS target that only affect
        with-tools: verilator'''
        self.chdir()
        logfile = 'target_with_additive_config_tools_test.log'
        rc = self.log_it('sim --tool verilator --debug target_with_additive_config_tools_test',
                         logfile=logfile)
        assert rc == 0
        # This target added to the Verilator waivers -Wno-style, -Wno-fatal,
        # but the defaults should also be there (at least -Wno-UNSIGNED)
        waivers = self.get_log_words_with('-Wno-', logfile=logfile)
        assert '-Wno-style' in waivers
        assert '-Wno-fatal' in waivers
        assert '-Wno-UNSIGNED' in waivers


    @pytest.mark.skipif('vivado' not in tools_loaded, reason="requires vivado")
    def test_tags_with_tools_vivado(self):
        '''test for DEPS target with tag using with-tools: verilator

        Since we're running --tool=vivado (not verilator) this should not
        apply the arg --lint-only in the DEPS tag, and instead run the
        full simulation.'''
        self.chdir()
        logfile = 'vivado_eda.log'
        rc = self.log_it('sim --tool vivado target_test', logfile=logfile)
        assert rc == 0

        # make sure the tag wasn't applied (should only be applied in verilator)
        # so the full sim should have run (xvlog, xelab, xsim) (--lint-only not applied,
        # b/c that should only apply in 'verilator' for this target.)
        exec_lines = self.get_log_lines_with('exec: ', logfile=logfile)
        assert len(exec_lines) == 3
        assert 'xvlog ' in exec_lines[0]
        assert 'xelab ' in exec_lines[1]
        assert 'xsim ' in exec_lines[2]
        assert not self.is_in_log('--lint-only', logfile=logfile)


    @pytest.mark.skipif('verilator' not in tools_loaded, reason="requires verilator")
    def test_tags_with_tools_add_incdirs(self):
        '''test for DEPS target with tag that adds incdirs'''
        self.chdir()
        logfile = 'target_foo_sv_add_incdirs.log'
        rc = self.log_it('elab --tool verilator target_foo_sv_add_incdirs',
                         logfile=logfile)
        assert rc == 0
        # This target added . to incdirs in the DEPS.yml dir.

        incdirs = self.get_log_words_with('+incdir+', logfile=logfile)
        assert len(incdirs) == 1 # should only have 1
        assert 'tests/deps_files/tags_with_tools' in incdirs[0]


    @pytest.mark.skipif('verilator' not in tools_loaded, reason="requires verilator")
    def test_tags_with_tools_add_defines(self):
        '''test for DEPS target with tag that adds defines'''
        self.chdir()
        logfile = 'target_foo_sv_add_defines.log'
        rc = self.log_it('elab --tool verilator --debug target_foo_sv_add_defines',
                         logfile=logfile)
        assert rc == 0
        assert self.is_in_log('+define+FOO_SV=3000', logfile=logfile)



'''opencos.commands.sweep - command handler for: eda sweep ...

These are not intended to be overriden by child classes. They do not inherit Tool classes.
'''

import os
import re

from opencos import util
from opencos.eda_base import CommandDesign, CommandParallel, get_eda_exec, which_tool

class CommandSweep(CommandDesign, CommandParallel):
    '''Command handler for: eda sweep ...'''

    command_name = 'sweep'

    def __init__(self, config:dict):
        CommandDesign.__init__(self, config=config)
        CommandParallel.__init__(self, config=config)
        self.sweep_target = ''
        self.single_command = ''
        self.args.update({
            'sweep': []
        })
        self.args_help.update({
            "sweep": ("List append arg, where range or value expansion is peformed on the RHS:"
                      " --sweep='--arg0=(start,last,iter=1)' "
                      " --sweep='--arg1=[val0,val1,val2,...]' "
                      " --sweep='+define+NAME0=(start,last,iter=1)' "
                      " --sweep='+define+NAME1=[val0,val1,val2,...]' "
                      " --sweep='+define+[NAME0,NAME1,NAME2,...]' ."
                      " Note that range expansion of (1,4) will expand to values [1,2,3,4]."
                      ),
        })


    def check_args(self) -> None:
        '''Returns None, checks self.args (use after args parsed)'''
        if self.args['parallel'] < 1 or self.args['parallel'] > 256:
            self.error(f"Arg {self.args['parallel']=} must be between 1 and 256")

    def _append_sweep_args(self, arg_tokens: list) -> None:
        '''Modifies list arg_tokens, using known top-level args'''
        tool = which_tool(command=self.single_command, config=self.config)
        super().update_args_list(args=arg_tokens, tool=tool)


    def process_tokens(
            self, tokens: list, process_all: bool = True,
            pwd: str = os.getcwd()
    ) -> list:
        '''CommandSweep.process_tokens(..) is likely the entry point for: eda sweep <command> ...

        - handles remaining CLI arguments (tokens list)
        - builds sweep_axis_list to run multiple jobs for the target
        '''

        # 'sweep' is special in the way it handles tokens, due to most of them being processed by
        # a sub instance
        sweep_axis_list = []
        arg_tokens = []

        _, unparsed = self.run_argparser_on_list(
            tokens=tokens,
            parser_arg_list=[
                'parallel',
                'sweep',
            ],
            apply_parsed_args=True
        )

        self.check_args()

        self.single_command = self.get_command_from_unparsed_args(tokens=unparsed)

        self._append_sweep_args(arg_tokens=arg_tokens)

        for sweep_arg_value in self.args['sweep']:
            # Deal with --sweep= args we already parsed, but haven't expanded yet.
            sweep_arg_value = util.strip_outer_quotes(sweep_arg_value)
            sweep_axis_list_entry = self._process_sweep_arg(sweep_arg_value=sweep_arg_value)
            if sweep_axis_list_entry:
                sweep_axis_list.append(sweep_axis_list_entry)


        # command, --parallel, and --sweep already processed by argparse,
        # let's tentatively parse what the child jobs cannot consume.
        # Whatever is leftover is either an eda target, or another unparsed token.
        single_cmd_unparsed = self.get_unparsed_args_on_single_command(
            command=self.single_command, tokens=unparsed
        )

        for token in unparsed:
            if token in single_cmd_unparsed:

                if self.resolve_target(token, no_recursion=True):
                    if self.sweep_target:
                        self.error("Sweep can only take one target, already got",
                                   f"{self.sweep_target} now getting {token}")
                    self.sweep_target = token
                else:
                    # If we don't know what to do with it, pass it to downstream
                    arg_tokens.append(token)

            else:
                # If it wasn't in single_cmd_unparsed, then it can definitely be
                # consumed downstream
                arg_tokens.append(token)


        # now we need to expand the target list
        util.debug(f"Sweep: command:    '{self.single_command}'")
        util.debug(f"Sweep: arg_tokens: '{arg_tokens}'")
        util.debug(f"Sweep: target:     '{self.sweep_target}'")

        # now create the list of jobs, support one axis
        self.jobs = []

        self.expand_sweep_axis(arg_tokens=arg_tokens, sweep_axis_list=sweep_axis_list)
        self.run_jobs(command=self.single_command)
        return [] # we used all of unparsed.


    @staticmethod
    def _process_sweep_arg(sweep_arg_value: str) -> dict:
        '''If processed, returns a non-empty dict. Handles processing of --sweep=VALUE

        Where VALUE could be one of:
           --arg0=(1,4)
           --arg1=[val0,val1]
           +define+NAME0=[val0,val1]

        Return value is {} or {'lhs': str, 'operator': str (+ or =), 'values': list}
        '''

        sweep_arg_value = util.strip_outer_quotes(sweep_arg_value)

        util.debug(f'{sweep_arg_value=}')
        # Try to match a sweep range expansion:
        #   --sweep='--arg0=(1,4)'
        #   --sweep='--arg0=(1,4,1)'
        #   --sweep='+define+NAME0=(1,4,1)'
        m = re.match(
            r'(.*)(\=) \(  ([\d\.]+)  \,  ([\d\.]+)  (\, ([\d\.]+) )?  \)'.replace(' ',''),
            sweep_arg_value
        )
        if m:
            lhs = m.group(1)
            operator = m.group(2)
            sweep_axis_values = []
            if m.group(5):
                rhs_range_iter = int(m.group(6))
            else:
                rhs_range_iter = 1
            for v in range(int(m.group(3)), int(m.group(4)) + 1, rhs_range_iter):
                sweep_axis_values.append(v)
            util.debug(f"Sweep axis: {lhs} {operator} {sweep_axis_values}")
            return {
                'lhs': lhs,
                'operator': operator,
                'values': sweep_axis_values,
            }

        # Try to match a sweep list value expansion:
        #   --sweep='--arg0=[1,2]'
        #   --sweep='--arg1=[3,4,5,6]'
        #   --sweep='+define+NAME=[8,9]'
        #   --sweep='+define+[NAME0,NAME1,NAME2]'
        #   --sweep='+define+[NAME0=1,NAME1=22,NAME2=46]'
        m = re.match(
            r'(.*)([\=\+]) \[  ([^\]]+) \] '.replace(' ',''),
            sweep_arg_value
        )
        if m:
            sweep_axis_values = []
            lhs = m.group(1)
            operator = m.group(2)
            for v in m.group(3).split(','):
                sweep_axis_values.append(v)
            util.debug(f"Sweep axis: {lhs} {operator} {sweep_axis_values}")
            return {
                'lhs': lhs,
                'operator': operator,
                'values': sweep_axis_values,
            }

        util.warning(f'Ignored unprocessed expansion for --sweep={sweep_arg_value}')
        return {}


    def expand_sweep_axis(
            self, arg_tokens: list, sweep_axis_list: list, sweep_string: str = ""
    ) -> None:
        '''Returns None, appends jobs to self.jobs to be run by CommandParallel.run_jobs(..)'''

        util.debug(f"Entering expand_sweep_axis: command={self.single_command},",
                   f"target={self.sweep_target}, arg_tokens={arg_tokens},",
                   f"sweep_axis_list={sweep_axis_list}")
        if not sweep_axis_list:
            # we aren't sweeping anything, create one job
            snapshot_name = self.sweep_target.replace('../','').replace('/','_') + sweep_string
            eda_path = get_eda_exec('sweep')
            self.jobs.append({
                'name' : snapshot_name,
                'index' : len(self.jobs),
                'command_list' : (
                    [eda_path, self.single_command, self.sweep_target,
                     '--job_name', snapshot_name] + arg_tokens
                )
            })
            return

        sweep_axis = sweep_axis_list.pop(0)
        lhs = sweep_axis['lhs']
        operator = sweep_axis['operator']

        lhs_trimmed = lhs.replace('-', '').replace('+', '').replace('=', '')

        for v in sweep_axis['values']:
            this_arg_tokens = arg_tokens.copy()
            this_arg_tokens.append(f'{lhs}{operator}{v}')

            v_string = f"{v}".replace('.','p')
            this_sweep_string = sweep_string + f"_{lhs_trimmed}_{v_string}"

            self.expand_sweep_axis(
                arg_tokens=this_arg_tokens,
                sweep_axis_list=sweep_axis_list,
                sweep_string=this_sweep_string
            )



''' opencos.tools.slang - Used by opencos.eda for elab commands w/ --tool=slang

Contains classes for ToolSlang, CommandElabSlang
'''

# pylint: disable=R0801 # (calling functions with same arguments)

import os
import shutil
import subprocess

from opencos import util
from opencos.util import sanitize_defines_for_sh
from opencos.eda_base import Tool
from opencos.commands import CommandElab


class ToolSlang(Tool):
    '''ToolSlang used by opencos.eda for --tool=slang'''

    _TOOL = 'slang'
    _EXE = 'slang'
    _URL = 'https://github.com/MikePopoloski/slang'

    slang_exe = ''
    slang_base_path = ''
    slang_tidy_exe = ''
    slang_hier_exe = ''

    def get_versions(self) -> str:
        if self._VERSION:
            return self._VERSION
        path = shutil.which(self._EXE)
        if not path:
            self.error(f'"{self._EXE}" not in path, need to get it ({self._URL}')
        else:
            self.slang_exe = path
            self.slang_base_path, _ = os.path.split(path)
            self.slang_tidy_exe = os.path.join(self.slang_base_path, 'slang-tidy')
            self.slang_hier_exe = os.path.join(self.slang_base_path, 'slang-hier')

        version_ret = subprocess.run(
            [self.slang_exe, '--version'],
            capture_output=True,
            check=False
        )
        stdout = version_ret.stdout.decode('utf-8')
        util.debug(f'{path=} {version_ret=}')
        words = stdout.split() # slang version 8.0.6+b4a74b00
        if len(words) < 3:
            self.error(f'{self.slang_exe} --version: returned unexpected string {version_ret=}')
        version = words[2]
        left, _ = version.split('+')
        ver_list = left.split('.')
        if len(ver_list) != 3:
            self.error(f'{self.slang_exe} --version: returned unexpected string',
                       f'{version_ret=} {version=}')
        self._VERSION = left
        return self._VERSION

    def set_tool_defines(self):
        self.defines['OC_TOOL_SLANG'] = None # add define
        if 'SYNTHESIS' not in self.defines:
            self.defines['SIMULATION'] = None # add define
        # Expected to manually add SYNTHESIS command line or target, otherwise.
        # Similarly, you could use --tool slang_yosys for a synthesis friendly
        # elab in Yosys.


class CommandElabSlang(CommandElab, ToolSlang):
    '''CommandElabSlang is a command handler for: eda elab --tool=slang'''

    def __init__(self, config:dict):
        CommandElab.__init__(self, config=config)
        ToolSlang.__init__(self, config=self.config)
        self.args.update({
            'slang-args': [], # aka, --single-unit, --ast-json <fname>, --ast-json-source-info
            'slang-json': False, # sets all the args I know of for AST.
            'slang-top': '',
            'tidy': False, # run slang-tidy instead of slang
            'hier': False, # run slang-hier instead of slang
        })

        self.all_json_args = [
            '--ast-json', ## needs filename: slang.json'
            '--ast-json-source-info',
            '--ast-json-detailed-types',
        ]

        self.args_help.update({
            'tidy': "Runs 'slang-tidy' instead of 'slang', with no ast- args.",
            'hier': "Runs 'slang-hier' instead of 'slang', with no ast- args.",
        })


        # If we're in elab, so not in general ToolSlang, set define for SLANG
        self.defines.update({
            'SLANG': 1
        })

        self.slang_command_lists = []


    # Note that we follow parent class CommandSim's do_it() flow, that way --export args
    # are handled.
    def prepare_compile(self):
        ''' prepare_compile() - following parent Commandsim's run() flow'''
        self.set_tool_defines()
        self.write_eda_config_and_args()

        self.slang_command_lists = self.get_compile_command_lists()
        self.write_slang_sh()

    def compile(self):
        pass

    def elaborate(self):
        ''' elaborate() - following parent Commandsim's run() flow, runs slang_command_lists'''
        if self.args['stop-before-compile'] or \
           self.args['stop-after-compile']:
            return
        self.run_commands_check_logs(self.slang_command_lists)

    def get_compile_command_lists(self, **kwargs) -> list:
        '''Returns list of util.ShellCommandList, for slang we'll run this in elaborate()'''

        command_list = self._get_slang_command_list_start() # slang vs slang-tidy vs slang-hier
        command_list += self.tool_config.get('compile-args', '--single-unit').split()
        command_list += self.args['slang-args'] # add user args.
        command_list += self._get_slang_json_args(command_exe=command_list[0])

        # incdirs
        for value in self.incdirs:
            command_list += [ '--include-directory', value ]

        # defines:
        for k,v in self.defines.items():
            command_list.append( '--define-macro' )
            if v is None:
                command_list.append( k )
            else:
                # Generally we should only support int and str python types passed as
                # --define-macro {k}={v}
                command_list.append( f'{k}={sanitize_defines_for_sh(v)}' )

        # Because many elab target-name won't match the --top needed for
        # slang, we'll leave this to arg --slang-top:
        if self.args.get('slang-top', None):
            command_list += [ '--top', self.args['slang-top'] ]


        command_list += self.files_sv + self.files_v

        command_list = util.ShellCommandList(command_list, tee_fpath='compile.log')
        return [command_list]

    def write_slang_sh(self):
        '''Writes run_slang.sh to work-dir'''
        util.write_shell_command_file(
            dirpath=self.args['work-dir'],
            filename='run_slang.sh',
            command_lists=self.slang_command_lists,
            line_breaks=True
        )

    def get_elaborate_command_lists(self, **kwargs) -> list:
        # all handled by get_compile_command_lists()
        return []

    def get_post_simulate_command_lists(self, **kwargs) -> list:
        return []

    def _get_slang_command_list_start(self) -> list:
        command_list = [self.slang_exe]

        if self.args['tidy']:
            if not shutil.which(self.slang_tidy_exe):
                util.warning("Running tool slang with --tidy, but 'slang-tidy'",
                             "not in PATH, using 'slang' instead")
            else:
                command_list = [self.slang_tidy_exe]

        if self.args['hier']:
            if self.args['tidy']:
                util.warning('Running with --tidy and --heir, will attempt to use slang-hier')
            elif not shutil.which(self.slang_hier_exe):
                util.warning("Running tool slang with --hier, but 'slang-hier'",
                             "not in PATH, using 'slang' instead")
            else:
                command_list = [self.slang_hier_exe]

        return command_list

    def _get_slang_json_args(self, command_exe:str) -> list:
        command_list = []

        if self.args.get('slang-json', False) and command_exe == 'slang':
            for arg in self.all_json_args:
                if arg not in command_list:
                    command_list.append(arg)
                    if arg == '--ast-json': # needs filename
                        command_list.append('slang.json')

        return command_list



''' opencos.tools.yosys - base class for slang_yosys.py, invio_yosys.py, tabbycad_yosys.py

Contains classes for ToolYosys
'''

# pylint: disable=R0801 # (calling functions with same arguments)

import os
import shutil
import subprocess

from opencos import util
from opencos.eda_base import Tool
from opencos.commands import CommandSynth

class ToolYosys(Tool):
    '''Parent class for ToolTabbyCadYosys, ToolInvioYosys, ToolSlangYosys'''

    _TOOL = 'yosys'
    _EXE = 'yosys'
    _URL = 'https://yosyshq.readthedocs.io/en/latest/'

    yosys_exe = ''
    sta_exe = ''
    sta_version = ''

    def get_versions(self) -> str:
        if self._VERSION:
            return self._VERSION

        path = shutil.which(self._EXE)
        if not path:
            self.error(f'"{self._EXE}" not in path or not installed, see {self._URL}')
        else:
            self.yosys_exe = path

        # Unforunately we don't have a non-PATH friendly support on self._EXE to set
        # where standalone 'sta' is. Even though Yosys has 'sta' internally, Yosys does
        # not fully support timing constraints or .sdc files, so we have to run 'sta'
        # standalone.
        sta_path = shutil.which('sta')
        if sta_path:
            util.debug(f'Also located "sta" via {sta_path}')
            self.sta_exe = sta_path
            sta_version_ret = subprocess.run(
                [self.sta_exe, '-version'], capture_output=True, check=False
            )
            util.debug(f'{self.yosys_exe} {sta_version_ret=}')
            sta_ver = sta_version_ret.stdout.decode('utf-8').split()[0]
            if sta_ver:
                self.sta_version = sta_ver

        version_ret = subprocess.run(
            [self.yosys_exe, '--version'], capture_output=True, check=False
        )
        util.debug(f'{self.yosys_exe} {version_ret=}')

        # Yosys 0.48 (git sha1 aaa534749, clang++ 14.0.0-1ubuntu1.1 -fPIC -O3)
        words = version_ret.stdout.decode('utf-8').split()

        if len(words) < 2:
            self.error(f'{self.yosys_exe} --version: returned unexpected str {version_ret=}')
        self._VERSION = words[1]
        return self._VERSION

    def set_tool_defines(self):
        self.defines.update({
            'OC_TOOL_YOSYS': None
        })
        if 'OC_LIBRARY' not in self.defines:
            self.defines.update({
                'OC_LIBRARY_BEHAVIORAL': None,
                'OC_LIBRARY': "0"
            })


class CommonSynthYosys(CommandSynth, ToolYosys):
    '''Common parent class used by invio_yosys and tabbycad_yosys

    for child classes: CommandSynthInvioYosys and tabbycad_yosys.CommandSynthTabbyCadYosys
    '''

    def __init__(self, config:dict):
        CommandSynth.__init__(self, config=config)
        ToolYosys.__init__(self, config=self.config)

        self.args.update({
            'sta': False,
            'liberty-file': '',
            'sdc-file': '',
            'yosys-synth': 'synth',              # synth_xilinx, synth_altera, etc (see: yosys help)
            'yosys-pre-synth': ['prep', 'proc'], # command run in yosys prior to yosys-synth.
            'yosys-blackbox': [],                # list of modules that yosys will blackbox.
        })
        self.args_help.update({
            'sta': 'After running Yosys, run "sta" with --liberty-file.' \
            + ' sta can be installed via: https://github.com/The-OpenROAD-Project/OpenSTA',
            'sdc-file': '.sdc file to use with --sta, if not present will use auto constraints',
            'liberty-file': 'Single liberty file for synthesis and sta,' \
            + ' for example: github/OpenSTA/examples/nangate45_slow.lib.gz',
            'yosys-synth': 'The synth command provided to Yosys, see: yosys help.',
            'yosys-pre-synth': 'Yosys commands performed prior to running "synth"' \
            + ' (or eda arg value for --yosys-synth)',
            'yosys-blackbox': 'List of modules that yosys will blackbox, likely will need these' \
            + ' in Verilog-2001 for yosys to read outside of slang and synth',
        })

        self.yosys_out_dir = ''
        self.yosys_v_path = ''
        self.full_work_dir = ''
        self.blackbox_list = []

    def do_it(self) -> None:
        self.set_tool_defines()
        self.write_eda_config_and_args()

        # Set up some dirs and filenames.
        self.full_work_dir = self.args.get('work-dir', '')
        if not self.full_work_dir:
            self.error(f'work_dir={self.full_work_dir} is not set')
        self.full_work_dir = os.path.abspath(self.full_work_dir)
        self.yosys_out_dir = os.path.join(self.full_work_dir, 'yosys')
        util.safe_mkdir(self.yosys_out_dir)
        self.yosys_v_path = os.path.join(self.yosys_out_dir, f'{self.args["top"]}.v')

        if self.is_export_enabled():
            self.do_export()
            return

        self.write_and_run_yosys_f_files()

    def write_and_run_yosys_f_files(self, **kwargs) -> None:
        '''Derived classes must define, to run remainder of do_it() steps'''
        raise NotImplementedError


    def create_yosys_synth_f(self) -> util.ShellCommandList:
        '''Derived classes may define, if they wish to get a list of yosys commands'''
        return util.ShellCommandList([])


    def get_synth_command_lines(self) -> list:
        '''Common yosys tcl after all blackbox and read_verilog commands'''

        lines = []
        lines += self.args.get('yosys-pre-synth', [])

        synth_command = self.args.get('yosys-synth', 'synth')
        if self.args['flatten-all']:
            synth_command += ' -flatten'

        lines.append(synth_command)


        # TODO(drew): I need a blackbox flow here? Or a memory_libmap?
        #   --> https://yosyshq.readthedocs.io/projects/yosys/en/latest/cmd/memory_libmap.html
        # TODO(drew): can I run multiple liberty files?
        if self.args['liberty-file']:
            lines += [
                'dfflibmap -liberty ' + self.args['liberty-file'],
                #'memory_libmap -lib ' + self.args['liberty-file'], # Has to be unzipped?
                'abc -liberty  ' + self.args['liberty-file'],
            ]
        lines += [
            'opt_clean',
            f'write_verilog {self.yosys_v_path}',
            f'write_json {self.yosys_v_path}.json',
        ]
        return lines


    def create_sta_f(self) -> util.ShellCommandList:
        '''Returns command list, for running 'sta' on sta.f'''

        if not self.args['sta']:
            return []

        if not self.args['liberty-file']:
            self.error('--sta is set, but need to also set --liberty-file=<file>')

        if self.args['sdc-file']:
            if not os.path.exists(self.args['sdc-file']):
                self.error(f'--sdc-file={self.args["sdc-file"]} file does not exist')

        if not self.sta_exe:
            self.error(f'--sta is set, but "sta" was not found in PATH, see: {self._URL}')

        sta_command_list = util.ShellCommandList(
            [ self.sta_exe, '-no_init', '-exit', 'sta.f' ],
            tee_fpath = 'sta.log'
        )

        # Need to create sta.f:
        if self.args['sdc-file']:
            sdc_path = self.args['sdc-file']
        else:
            # Need to create sdc.f:
            sdc_path = 'sdc.f'
            self.create_sdc_f()

        with open(os.path.join(self.args['work-dir'], 'sta.f'), 'w',
                  encoding='utf-8') as f:
            lines = [
                'read_liberty ' + self.args['liberty-file'],
                'read_verilog ' + self.yosys_v_path,
                'link_design ' + self.args['top'],
                'read_sdc ' + sdc_path,
                'report_checks',
            ]
            f.write('\n'.join(lines))

        return util.ShellCommandList(
            sta_command_list,
            tee_fpath = 'sta.log'
        )


    def create_sdc_f(self) -> None:
        '''Returns None, creates sdc.f'''

        if self.args['sdc-file']:
            # already exists from args, return b/c nothing to create.
            return

        with open(os.path.join(self.args['work-dir'], 'sdc.f'), 'w',
                  encoding='utf-8') as f:
            clock_name = self.args['clock-name']
            period = self.args['clock-ns']
            name_not_equal_clocks_str = f'NAME !~ "{clock_name}"'
            lines = [
                f'create_clock -add -name {clock_name} -period {period} [get_ports ' \
                + '{' + clock_name + '}];',
                f'set_input_delay -max {self.args["idelay-ns"]} -clock {clock_name}' \
                + ' [get_ports * -filter {DIRECTION == IN && ' \
                + name_not_equal_clocks_str + '}];',
                f'set_output_delay -max {self.args["odelay-ns"]} -clock {clock_name}' \
                + ' [get_ports * -filter {DIRECTION == OUT}];',
            ]
            f.write('\n'.join(lines))



#!/usr/bin/python3

# SPDX-License-Identifier: MPL-2.0

''' names.py

This is a human-maintained but ideally machine parsable mapping of names

Where convenient, this file should be parsed versus embedding names in code,
which is what oc_cli.py does.

Once assigned, names are not changed, so it's OK to use: `define OC_VENDOR 1
to indicate Xilinx in some RTL code.  There doesn't yet seem to be cause to
bring in automation to create SystemVerilog defines and such, as each board
just needs to set a couple of these statically.

all values in this file are in hex.
'''

table = {

    'OC_VENDOR': {
        0: "None",
        1: "Xilinx",
    },

    'OC_BOARD': {
        0: "None",
        1: "JC35",
        2: "U200",
        3: "U50",
        4: "U55N",
        5: "U50C",
        6: "PYNQ-Z2",
    },

    'OC_LIBRARY': {
        0: "None",
        1: "Ultrascale+",
    },

    'PLL_TYPES': {
        0: "None",
        1: "MMCME4_ADV",
    },

    'CHIPMON_TYPES': {
        0: "None",
        1: "SYSMONE4",
    },

    'IIC_OFFLOAD_TYPES': {
        0: "None",
        1: "Xilinx AXI-IIC",
    },

    'PCIE_TYPES': {
        0: "None",
        1: "Xilinx PCIE40E4--based XDMA",
    },

}



'''opencos.commands.synth - Base class command handler for: eda synth ...

Intended to be overriden by Tool based classes (such as CommandSynthVivado, etc)
'''

# Note - similar code waiver, tricky to eliminate it with inheritance when
# calling reusable methods.
# pylint: disable=R0801

import os

from opencos import util, export_helper
from opencos.eda_base import CommandDesign, Tool


class CommandSynth(CommandDesign):
    '''Base class command handler for: eda synth ...'''

    CHECK_REQUIRES = [Tool]
    error_on_no_files_or_targets = True
    error_on_missing_top = True

    command_name = 'synth'

    def __init__(self, config: dict):
        CommandDesign.__init__(self, config=config)
        self.args.update({
            'flatten-all': False,
            'flatten-none':  False,
            'clock-name': 'clock',
            'clock-ns': 5,
            'idelay-ns': 2,
            'odelay-ns': 2,
            'synth-blackbox': [],
        })
        self.defines['SYNTHESIS'] = None

    def do_it(self) -> None:
        '''Common do_it() method that child classes can use prior to customization'''

        # set_tool_defines() is from class Tool. Since that is not inherited yet, but
        # should be by any handlers like CommandSynthSlang, etc, check on the existence
        # of set_tool_defines, and error if not present.
        if not all(isinstance(self, x) for x in self.CHECK_REQUIRES):
            self.error('CommandSynth.do_it() requires a Tool to be in parent classes, but none is.',
                       f'{self.CHECK_REQUIRES=}')
            return

        # add defines for this job from Tool class if present
        self.command_safe_set_tool_defines() # (Command.command_safe_set_tool_defines)

        # dump our config to work-dir for debug
        self.write_eda_config_and_args()

        # optionally export
        if self.is_export_enabled():
            self.do_export()

        # Derived classes can do the rest, can call CommandSynth.do_it(self) as a first step.

    def process_tokens(self, tokens: list, process_all: bool = True,
                       pwd: str = os.getcwd()) -> list:
        unparsed = CommandDesign.process_tokens(
            self, tokens=tokens, process_all=process_all, pwd=pwd
        )

        if self.status_any_error():
            return unparsed

        # add defines for this job type
        if self.args['top']:
            # create our work dir
            self.create_work_dir()
            self.run_dep_commands()
            self.do_it()
        return unparsed

    def do_export(self):
        '''CommandSynth helper for handling args --export*

        We allow commands such as: eda synth --export <target>
        '''

        out_dir = os.path.join(self.args['work-dir'], 'export')

        target = self.target
        if not target:
            target = 'test'

        export_obj = export_helper.ExportHelper(
            cmd_design_obj=self,
            eda_command=self.command_name,
            out_dir=out_dir,
            # Note this may not be the correct target for debug infomation,
            # so we'll only have the first one.
            target=target
        )

        # Set things in the exported: DEPS.yml
        tool = self.args.get('tool', None)
        # Certain args are allow-listed here
        deps_file_args = []
        for a in self.get_command_line_args():
            if any(a.startswith(x) for x in [
                    '--optimize',
                    '--synth',
                    '--idelay',
                    '--odelay',
                    '--flatten',
                    '--clock',
                    '--yosys']):
                deps_file_args.append(a)

        export_obj.run(
            deps_file_args=deps_file_args,
            export_json_eda_config={
                'tool': tool,
            }
        )

        if self.args['export-run']:

            # remove the '--export' named args, we don't want those.
            args_no_export = self.get_command_line_args(remove_args_startswith=['export'])

            command_list = ['eda', self.command_name] + args_no_export + [target]

            util.info(f'export-run: from {export_obj.out_dir=}: {command_list=}')
            self.exec(
                work_dir=export_obj.out_dir,
                command_list=command_list,
            )



'''pytest functions for making sure the build has expected members (version, etc)'''

import opencos

def test_version():
    '''tests that opencos.__version__ works'''
    __version__ = opencos.__version__
    print(f'{__version__=}')
    assert __version__
    assert __version__ != 'unknown'
    numbers = __version__.split('.')
    assert any(int(number) != 0 for number in numbers)



''' opencos.tools.invio_helpers - general helper functions for running Invio.'''

# TODO(drew): clean these up, method write_py_file is too long, etc:
# pylint: disable=dangerous-default-value # default list arg = []
# pylint: disable=too-many-locals
# pylint: disable=too-many-branches
# pylint: disable=too-many-statements
# pylint: disable=unused-argument


import os
from typing import TypedDict
from opencos.util import ShellCommandList

# Given what we'll assume is a CommandDesign (CommandSim?) object, we'll try to
# perform some invio related commands on it.

class InvioCommandInfo(TypedDict):
    '''Return dict schema from get_invio_command_dict(..)'''
    work_dir: str
    py_filename: str
    v_filename: str
    full_v_filename: str
    full_py_filename: str
    top: str
    blackbox_module_list: list
    blackbox_files_list: list
    log_fname: str
    command_lists: list # list of util.ShellCommandList


def write_py_file(
        command_design_obj: object,
        py_filename: str = 'run_invio.py',
        v_filename: str = '',
        blackbox_list: list = [],
        sim_elab: bool = False,
        **kwargs
) -> dict:
    '''Given an eda.CommandDesign object, creates a .py file that can be run

    from invio-py or standalone. Returns a dict with a few details about the generated .py'''


    # Sanity check that command_design_obj has some expected details.
    assert isinstance(getattr(command_design_obj, 'incdirs', None), list)
    assert isinstance(getattr(command_design_obj, 'files_v', None), list)
    assert isinstance(getattr(command_design_obj, 'files_sv', None), list)
    assert isinstance(getattr(command_design_obj, 'files_vhd', None), list)

    args = getattr(command_design_obj, 'args', {})
    work_dir = args.get('work-dir', '')

    assert work_dir

    work_dir = os.path.abspath(work_dir)

    lines = [

        'from invio import init, define_macro, add_include_directory, \\',
        '    add_verilog_file, add_sv_file, elaborate, elaborate_pct, analyze, \\',
        '    print_instance_hierarchy, write_design, report_analyzed_files',
        'import os, shutil',
        '',
        'for p in ["invio"]:',
        f'    fullp = os.path.join("{work_dir}", p)',
        '    if not os.path.exists(fullp):',
        '        os.mkdir(fullp)',
        '',
    ]

    for name,value in command_design_obj.defines.items():
        if name in ['SIMULATION']:
            continue

        if value is None:
            lines.append(f'define_macro("{name}")')
        else:
            value = str(value)
            value = value.replace("'", "\\" + "'")
            value = value.replace('"', "\\" + '"')
            lines.append(f'define_macro("{name}", value="{value}")')

    # We must define SYNTHESIS for oclib_defines.vh to work correctly.
    lines.append('define_macro("SYNTHESIS")')

    for path in command_design_obj.incdirs:
        lines.append(f'add_include_directory("{path}")')

    for path in command_design_obj.files_v:
        lines.append(f'add_verilog_file("{path}")')

    for path in command_design_obj.files_sv:
        lines.append(f'add_sv_file("{path}")')

    for path in command_design_obj.files_vhd:
        lines.append(f'add_vhdl_file("{path}")')

    # Note - big assumption here that "module myname" is contained in myname.[v|sv]:
    blackbox_files_list = []
    for path in command_design_obj.files_v + command_design_obj.files_sv:
        _, leaf_filename = os.path.split(path)
        module_name = ''.join(leaf_filename.split('.')[:-1])
        if module_name in blackbox_list:
            blackbox_files_list.append(path)


    top = args.get('top', None)
    if top is None:
        # Infer the 'top' module as last verilog, SV, or vhdl file.
        tmp_list = command_design_obj.files_v + command_design_obj.files_sv \
            + command_design_obj.files_vhd
        _, top_fileonly = os.path.split(tmp_list[-1])
        top = ''.join(top_fileonly.split('.')[:-1])


    assert top

    if not v_filename:
        v_filename = f'{top}.v'


    full_v_filename = os.path.join(work_dir, 'invio', v_filename)
    full_py_filename = os.path.join(work_dir, py_filename)

    ret = InvioCommandInfo({
        'work_dir': work_dir,
        'py_filename': py_filename,
        'v_filename': v_filename,
        'full_v_filename': full_v_filename,
        'full_py_filename': full_py_filename,
        'top': top,
        'blackbox_module_list': blackbox_list,
        'blackbox_files_list': blackbox_files_list,
        'log_fname': 'invio.log',
        'command_lists': [ ShellCommandList(['python3', full_py_filename],
                                            tee_fpath = 'invio.log')]
    })

    if sim_elab:
        # elab (non-synthesis), runs the following (from eda.CommandElabInvio)
        lines += [
            'assert analyze()',
            f"assert elaborate('{top}', pct_elaboration=True, forceBlackbox={blackbox_list})",
            'assert elaborate_pct()',
            '',
            'report_analyzed_files()',
            'print_instance_hierarchy()',
        ]
    else:
        # synthesis-style elab (from eda.CommandElabInvioYosys)
        lines += [
            'assert analyze()',
            f"assert elaborate('{top}', rtl_elaboration=True, forceBlackbox={blackbox_list})",
            '',
            'report_analyzed_files()',
            'print_instance_hierarchy()',
            '',
            f"assert write_design('{full_v_filename}')",
            '',
        ]

        # It seems that Invio's write_design(..) will leave a few VERIFIC_MUX and other modules
        # commented out in the .v file. This doesn't work downstream for yosys, so uncomment these
        # with some brute force.

        tmp_v_filename = os.path.join(work_dir, '_tmp.v')

        lines += [
            '',
            '# We also get the joy of uncommenting any VERIFIC_* modules in the generated .v',
            f'with open("{full_v_filename}") as old, open("{tmp_v_filename}", "w") as new:',
            '    in_verific_commented_module = False',
            '    for line in old.readlines():',
            '        # Handle old, modify line:',
            '        if "// module VERIFIC_" in line:',
            '            in_verific_commented_module = True',
            '        if in_verific_commented_module:',
            '            line = line.replace("// ", "")',
            '        if line.startswith("endmodule"):',
            '            in_verific_commented_module = False',
            '        # Write out line to new:',
            '        new.write(line)',
            f'shutil.move("{tmp_v_filename}", ',
            f'            "{full_v_filename}")',
            '',
        ]


        for k,v in ret.items():
            if isinstance(v, str):
                lines.append(f'## {k} = "{v}"')

        # done with (not sim_elab).


    # write 'run_invo.py':
    if py_filename:
        outfile = os.path.join(work_dir, py_filename)
        with open(outfile, 'w', encoding='utf-8') as f:
            f.write('\n'.join(lines))


    return ret




def get_invio_command_dict(command_design_obj:object, v_filename='', py_filename='run_invio.py',
                           blackbox_list=[], sim_elab:bool=False, **kwargs) -> dict:
    '''Creates a .py file from an eda.CommandDesign. Returns a dict with

    information about what to run.'''

    invio_dict = write_py_file( command_design_obj, v_filename=v_filename,
                                py_filename=py_filename, blackbox_list=blackbox_list,
                                sim_elab=sim_elab, **kwargs )

    return invio_dict



'''opencos.tests.helpers - methods used to assist in various pytests on opencos'''

# pylint: disable=dangerous-default-value  # list arg w/ default values.

import json
import os
import shutil

from contextlib import redirect_stdout, redirect_stderr

from opencos import eda, util
from opencos import deps_schema

def can_run_eda_command(*commands, config: dict) -> bool:
    '''Returns True if we have any installed tool that can run: eda <command>'''
    runnable = []
    for command in list(commands):
        handler = config['command_handler'].get(command, None)
        if not handler:
            return False
        if handler and getattr(handler, 'CHECK_REQUIRES', []):
            if not all(isinstance(handler, x) for x in getattr(handler, 'CHECK_REQUIRES', [])):
                return False
        runnable.append(True)
    return runnable and all(runnable)

def chdir_remove_work_dir(startpath, relpath):
    '''Changes dir to startpath/relpath, removes the work directories (eda.work, eda.export*)'''
    os.chdir(os.path.join(startpath, relpath))
    for outdir in ['eda.export', 'eda.work']:
        fullp = os.path.join(os.getcwd(), outdir)
        if fullp and ('eda.' in fullp) and os.path.isdir(fullp):
            shutil.rmtree(fullp)

def eda_wrap(*args):
    '''Calls eda.main, prefer seed=1 to avoid seed based simulation fails in pytests'''
    main_args = [x for x in list(args) if '--seed' not in x]
    return eda.main('--seed=1', *main_args)

def eda_sim_wrap(*args):
    '''Calls eda.main for 'sim' prefer seed=1 to avoid seed based simulation fails in pytests'''
    main_args = [x for x in list(args) if (x != 'sim' and '--seed' not in x)]
    return eda.main('sim', '--seed=1', *main_args)

def eda_elab_wrap(*args):
    '''Calls eda.main for 'elab'.'''
    main_args = [x for x in list(args) if (x != 'elab' and '--seed' not in x)]
    return eda.main('elab', *main_args)

def assert_sim_log_passes(
        filepath: str, want_str: str = 'TEST PASS',
        err_strs: list = ['Error', 'ERROR', 'TEST FAIL']
) -> None:
    '''Checks log in filepath, makes sure it has a want_str and no err_strs'''
    test_passed = False
    test_failed = False

    assert os.path.exists(filepath), f'{filepath=} does not exist'
    if not want_str:
        # we don't have a want_str, so looks like it passes no matter what
        test_passed = True
    with open(filepath, encoding='utf-8') as f:
        for line in f.readlines():
            if want_str and want_str in line:
                test_passed = True
            if any(x in line for x in err_strs):
                test_failed = True
    assert test_passed, f'{filepath=} did not have {want_str=}'
    assert not test_failed, f'{filepath=} has one of {err_strs=}'

def assert_gen_deps_yml_good(filepath:str, want_target:str='') -> dict:
    '''Generated DEPS files should be coming from --export style args,

    so we also confirm they pass the deps_schema.FILE_SIMPLIFIED'''
    assert os.path.exists(filepath), f'{filepath=} does not exist'
    data = util.yaml_safe_load(filepath)
    assert len(data.keys()) > 0
    if want_target:
        assert want_target, f'{want_target=} not in {filepath=} {data=}'
        assert 'deps' in data[want_target], f' key "deps" is not in {want_target=} in {data=}'
    assert deps_schema.check_files(filepath, schema_obj=deps_schema.FILE_SIMPLIFIED)
    return data


def assert_export_json_good(filepath:str) -> dict:
    '''Checks that an exported JSON (from eda export, or eda <command> --export) has known keys'''
    assert os.path.exists(filepath), f'{filepath=} does not exist'
    with open(filepath, encoding='utf-8') as f:
        data = json.load(f)
    assert 'name' in data
    assert 'eda' in data
    assert any(x in data for x in ['files', 'tb'])
    return data


def assert_export_jsonl_good(filepath:str, jsonl:bool=True) -> list:
    '''Checks that an exported JSONL (from eda multi --export) has known keys'''
    assert os.path.exists(filepath), f'{filepath=} does not exist'
    ret = []
    with open(filepath, encoding='utf-8') as f:
        if jsonl:
            for line in f.readlines():
                line = line.strip()
                data = json.loads(line)
                assert 'name' in data
                assert 'eda' in data
                assert any(x in data for x in ['files', 'tb'])
                ret.append(data)
        else:
            data = json.load(f)
            assert 'tests' in data
            assert isinstance(data['tests'], list)
            for entry in data['tests']:
                assert 'name' in entry
                assert 'eda' in entry
                assert any(x in entry for x in ['files', 'tb'])
                ret.append(entry)


    return ret


class Helpers:
    '''We do so much with logging in this file, might as well make it reusable'''
    DEFAULT_DIR = ''
    DEFAULT_LOG = 'eda.log'
    def chdir(self):
        '''Changes directory to self.DEFAULT_DIR and removes eda.work, eda.export paths'''
        chdir_remove_work_dir('', self.DEFAULT_DIR)

    def log_it(self, command_str:str, logfile=None, use_eda_wrap=True):
        '''Replacement for calling eda.main or eda_wrap, when you want an internal logfile

        Usage:
            rc = self.log_it('sim foo', logfile='yes.log')
            assert rc == 0
        '''
        if logfile is None:
            logfile = self.DEFAULT_LOG
        rc = 50

        # TODO(drew): There are some issues with log_it redirecting stdout from vivado
        # and modelsim_ase. So this may not work for all tools, you may have to directly
        # look at eda.work/{target}.sim/sim.log or xsim.log.
        print(f'{os.getcwd()=}')
        print(f'{command_str=}')
        with open(logfile, 'w', encoding='utf-8') as f:
            with redirect_stdout(f), redirect_stderr(f):
                if use_eda_wrap:
                    rc = eda_wrap(*(command_str.split()))
                else:
                    rc = eda.main(*(command_str.split()))
            print(f'Wrote: {os.path.abspath(logfile)=}')
        return rc

    def is_in_log(self, *want_str, logfile=None, windows_path_support=False):
        '''Check if any of want_str args are in the logfile, or self.DEFAULT_LOG'''
        if logfile is None:
            logfile = self.DEFAULT_LOG
        want_str0 = ' '.join(list(want_str))
        want_str1 = want_str0.replace('/', '\\')
        with open(logfile, encoding='utf-8') as f:
            for line in f.readlines():
                if want_str0 in line or \
                   (windows_path_support and want_str1 in line):
                    return True
        return False

    def get_log_lines_with(self, *want_str, logfile=None, windows_path_support=False):
        '''gets all log lines with any of want_str args are in the logfile, or self.DEFAULT_LOG'''
        if logfile is None:
            logfile = self.DEFAULT_LOG
        ret_list = []
        want_str0 = ' '.join(list(want_str))
        want_str1 = want_str0.replace('/', '\\')
        with open(logfile, encoding='utf-8') as f:
            for line in f.readlines():
                if want_str0 in line:
                    ret_list.append(line)
                elif windows_path_support and want_str1 in line:
                    ret_list.append(line)
        return ret_list

    def get_log_words_with(self, *want_str, logfile=None, windows_path_support=False):
        '''gets all log lines with any of *want_str within a single word
        in the logfile or self.DEFAULT_LOG
        '''
        if logfile is None:
            logfile = self.DEFAULT_LOG
        ret_list = []
        want_str0 = ' '.join(list(want_str))
        want_str1 = want_str0.replace('/', '\\')
        with open(logfile, encoding='utf-8') as f:
            for line in f.readlines():
                if want_str0 in line:
                    for word in line.split():
                        if want_str0 in word:
                            ret_list.append(word)
                elif windows_path_support and want_str1 in line:
                    for word in line.split():
                        if want_str1 in word:
                            ret_list.append(word)

        return ret_list



import os
import sys
import argparse
import mergedeep
import shutil

from opencos import util

class Defaults:
    # Order of precedence for default value of eda arg: --config-yaml:
    # 1) os.environ.get('EDA_CONFIG_YML', '') -- filepath to an eda_config.yml file
    # 2) ~/.opencos-eda/EDA_CONFIG.yml
    # 3) (package pip installed dir for opencos)/eda_config_defaults.yml
    environ_override_config_yml = os.environ.get('EDA_CONFIG_YML', '')
    home_override_config_yml = os.path.join(
        os.environ.get('HOME', ''), '.opencos-eda', 'EDA_CONFIG.yml'
    )
    opencos_config_yml = 'eda_config_defaults.yml'
    config_yml = ''

    supported_config_keys = set([
        'DEFAULT_HANDLERS',
        'defines',
        'dep_command_enables',
        'dep_tags_enables',
        'deps_markup_supported',
        'deps_subprocess_shell',
        'bare_plusarg_supported',
        'dep_sub',
        'vars',
        'file_extensions',
        'command_determines_tool',
        'command_uses_no_tools',
        'tools',
        'auto_tools_order',
    ])
    supported_config_auto_tools_order_keys = set([
        'exe', 'handlers', 'requires_env', 'requires_py', 'requires_cmd',
        'disable-tools-multi',
    ])
    supported_config_tool_keys = set([
        'defines',
        'log-bad-strings',
        'log-must-strings',
        'sim-libraries',
        'compile-args',
        'compile-waves-args',
        'compile-waivers',
        'compile-coverage-args',
        'elab-args',
        'elab-waves-args',
        'simulate-args',
        'simulate-waves-args',
        'simulate-waivers',
        'coverage-args',
    ])


if os.path.exists(Defaults.environ_override_config_yml):
    Defaults.config_yml = Defaults.environ_override_config_yml
elif os.path.exists(Defaults.home_override_config_yml):
    Defaults.config_yml = Defaults.home_override_config_yml
else:
    Defaults.config_yml = Defaults.opencos_config_yml


def find_eda_config_yml_fpath(filename:str, package_search_only=False, package_search_enabled=True) -> str:
    '''Locates the filename (.yml) either from fullpath provided or from the sys.path
    opencos package paths.'''

    # Check fullpath, unless we're only checking the installed pacakge dir.
    if package_search_only:
        pass
    elif os.path.exists(filename):
        return os.path.abspath(filename)

    leaf_filename = os.path.split(filename)[1]

    if leaf_filename != filename:
        # filename had subdirs, and we didn't find it already.
        util.error(f'eda_config: Could not find {filename=}')
        return None

    # Search in . or pacakge installed dir
    thispath = os.path.dirname(__file__) # this is not an executable, should be in packages dir.

    if package_search_only:
        paths = [thispath]
    elif package_search_enabled:
        paths = ['', thispath]
    else:
        paths = ['']


    for dpath in paths:
        fpath = os.path.join(dpath, leaf_filename)
        if os.path.exists(fpath):
            return fpath

    util.error(f'eda_config: Could not find {leaf_filename=} in opencos within {paths=}')
    return None


def check_config(config:dict, filename='') -> None:
    # sanity checks:
    for key in config:
        if key not in Defaults.supported_config_keys:
            util.error(f'eda_config.get_config({filename=}): has unsupported {key=}' \
                       + f' {Defaults.supported_config_keys=}')

    for row in config.get('auto_tools_order', []):
        for tool, table in row.items():
            for key in table:
                if key not in Defaults.supported_config_auto_tools_order_keys:
                    util.error(f'eda_config.get_config({filename=}): has unsupported {key=}' \
                               + f' in auto_tools_order, {tool=},' \
                               + f' {Defaults.supported_config_auto_tools_order_keys=}')

    for tool,table in config.get('tools', {}).items():
        for key in table:
            if key not in Defaults.supported_config_tool_keys:
                util.error(f'eda_config.get_config({filename=}): has unsupported {key=}' \
                           + f' in config.tools.{tool=}, ' \
                           + f' {Defaults.supported_config_tool_keys=}')


def update_config_auto_tool_order_for_tool(tool: str, config: dict) -> str:
    '''Update config entry if the value for tool is in the form 'name=/path/to/exe

    Input arg tool can be in the form (for example):
      tool='verlator', tool='verilator=/path/to/verilator.exe'

    Performs no update if tool has no = in it. Returns tool (str) w/out = in it
    '''
    if not tool or '=' not in tool:
        return tool

    tool, user_exe = tool.split('=')[0:2]

    user_exe = shutil.which(user_exe)

    if tool not in config['auto_tools_order'][0]:
        return tool
    if not user_exe:
        return tool

    old_exe = config['auto_tools_order'][0][tool].get('exe', str())
    if type(old_exe) is list:
        config['auto_tools_order'][0][tool]['exe'][0] = user_exe
    else:
        config['auto_tools_order'][0][tool]['exe'] = user_exe
    return tool


def update_config_auto_tool_order_for_tools(tools: list, config: dict) -> list:
    ret = list()
    for tool in tools:
        ret.append(update_config_auto_tool_order_for_tool(tool, config))
    return ret


def update_config_for_eda_safe(config) -> None:
    config['dep_command_enables']['shell'] = False


def deps_shell_commands_enabled(config) -> bool:
    return config['dep_command_enables']['shell']


def get_config(filename) -> dict:
    fpath = find_eda_config_yml_fpath(filename)
    user_config = util.yaml_safe_load(fpath)
    check_config(user_config, filename=filename)

    # The final thing we do is update key 'config-yml' with the full path used.
    # This way we don't have to pass around --config-yml as some special arg
    # in eda.CommandDesign.args, and eda.CommandMulti can use when re-invoking 'eda'.
    user_config['config-yml'] = fpath
    return user_config


def get_config_handle_defaults(filename) -> dict:
    user_config = get_config(filename)
    user_config = get_config_merged_with_defaults(user_config)
    return user_config


def merge_config(dst_config:dict, overrides_config:dict, additive_strategy=False) -> None:
    '''Mutates dst_config, uses Strategy.TYPESAFE_REPLACE'''
    # TODO(drew): It would be cool if I could have Sets be additive, but oh well,
    # this gives the user more control over replacing entire lists.
    strategy = mergedeep.Strategy.TYPESAFE_REPLACE
    if additive_strategy:
        strategy = mergedeep.Strategy.TYPESAFE_ADDITIVE
    mergedeep.merge(dst_config, overrides_config, strategy=strategy)


def get_config_merged_with_defaults(config:dict) -> dict:
    default_fpath = find_eda_config_yml_fpath(Defaults.config_yml, package_search_only=True)
    default_config = util.yaml_safe_load(default_fpath)
    merge_config(default_config, overrides_config=config)
    # This technically mutated updated into default_config, so return that one:
    return default_config

def get_argparser() -> argparse.ArgumentParser:
    parser = argparse.ArgumentParser(prog='opencos eda config options', add_help=False, allow_abbrev=False)
    parser.add_argument('--config-yml', type=str, default=Defaults.config_yml,
                        help=f'YAML filename to use for configuration (default {Defaults.config_yml})')
    return parser

def get_argparser_short_help() -> str:
    return util.get_argparser_short_help(parser=get_argparser())


def get_eda_config(args:list, quiet=False) -> (dict, list):
    '''Returns an config dict and a list of args to be passed downstream
    to eda.main and eda.process_tokens.

    Handles args for:
      --config-yml=<YAMLFILE>

    This will merge the result with the default config (if overriden)
    '''

    parser = get_argparser()
    try:
        parsed, unparsed = parser.parse_known_args(args + [''])
        unparsed = list(filter(None, unparsed))
    except argparse.ArgumentError:
        util.error(f'problem attempting to parse_known_args for {args=}')

    util.debug(f'eda_config.get_eda_config: {parsed=} {unparsed=}  from {args=}')

    if parsed.config_yml:
        if not quiet:
            util.info(f'eda_config: --config-yml={parsed.config_yml} observed')
        fullpath = find_eda_config_yml_fpath(parsed.config_yml)
        config = get_config(fullpath)
        if not quiet:
            util.info(f'eda_config: using config: {fullpath}')

        # Calling get_config(fullpath) will add fullpath to config['config-yml'], so the
        # arg for --config-yml does not need to be re-added.
    else:
        config = None

    if parsed.config_yml != Defaults.config_yml:
        config = get_config_merged_with_defaults(config)

    return config, unparsed



'''opencos.commands.elab - Base class command handler for: eda elab ...

Intended to be overriden by Tool based classes (such as CommandElabVivado, etc)

Note that many 'elab' command handlers that also can perform simulations, such
as VerilatorElab, will instead inherit from VerilatorSim and simply perform a
shortened sim as the elab, instead of inheriting CommandElab.

Tools that don't support a 'sim' command will generally use CommandElab, such
as CommandElabSlang.'''

from opencos.commands.sim import CommandSim

class CommandElab(CommandSim):
    '''Base class command handler for: eda elab ...'''

    command_name = 'elab'

    def __init__(self, config: dict):
        CommandSim.__init__(self, config=config)
        # add args specific to this simulator
        self.args['stop-after-elaborate'] = True
        self.args['lint'] = True
        self.args['verilate-args'] = []



    def compile(self) -> None:
        raise NotImplementedError

    def elaborate(self) -> None:
        raise NotImplementedError

    def get_compile_command_lists(self, **kwargs) -> list:
        ''' Returns a list of lists (list of command lists).'''
        raise NotImplementedError

    def get_elaborate_command_lists(self, **kwargs) -> list:
        ''' Returns a list of lists (list of command lists).'''
        raise NotImplementedError

    # CommandSim methods that elab does not use:

    def simulate(self):
        pass

    def get_simulate_command_lists(self, **kwargs) -> list:
        return []

    def get_post_simulate_command_lists(self, **kwargs) -> list:
        return []



''' opencos.tools.modelsim_ase - Used by opencos.eda for sim/elab commands w/ --tool=modelsim_ase.

Contains classes for ToolModelsimAse, CommandSimModelsimAse, CommandElabModelsimAse.
'''

import os

from opencos import util
from opencos.util import sanitize_defines_for_sh
from opencos.commands import CommandSim
from opencos.tools.questa import ToolQuesta

class ToolModelsimAse(ToolQuesta):
    '''ToolModelsimAse used by opencos.eda for --tool=modelsim_ase'''

    _TOOL = 'modelsim_ase' # otherwise it's 'questa' from base class.
    _EXE = 'vsim'

class CommandSimModelsimAse(CommandSim, ToolModelsimAse):
    '''CommandSimModelsimAse is a command handler for: eda sim --tool=modelsim_ase'''

    def __init__(self, config: dict):
        CommandSim.__init__(self, config=config)
        ToolModelsimAse.__init__(self, config=self.config)
        self.shell_command = os.path.join(self.sim_exe_base_path, 'vsim')
        self.starter_edition = True
        self.args.update({
            'tool': self._TOOL, # override
            'gui': False,
        })

    def set_tool_defines(self):
        # Update any defines from config.tools.modelsim_ase:
        self.defines.update(
            self.tool_config.get(
                'defines',
                # defaults, if not set:
                {'OC_ASSERT_PROPERTY_NOT_SUPPORTED': 1,
                 'OC_TOOL_MODELSIM_ASE': 1}
            )
        )

    # We do override do_it() to avoid using CommandSimQuesta.do_it()
    def do_it(self):
        CommandSim.do_it(self)
        #    self.compile()   # runs if stop-before-compile is False, stop-after-compile is True
        #    self.elaborate() # runs if stop-before-compile is False, stop-after-compile is False,
        #                       stop-after-elaborate is True
        #    self.simulate()  # runs if stop-* are all False (run the whole thing)


    def prepare_compile(self):
        self.set_tool_defines()
        self.write_vlog_dot_f()
        self.write_vsim_dot_do(dot_do_to_write='all')

        vsim_command_lists = self.get_compile_command_lists()
        util.write_shell_command_file(
            dirpath=self.args['work-dir'],
            filename='compile_only.sh',
            command_lists=vsim_command_lists
        )

        vsim_command_lists = self.get_elaborate_command_lists()
        util.write_shell_command_file(
            dirpath=self.args['work-dir'],
            filename='compile_elaborate_only.sh',
            command_lists=vsim_command_lists
        )

        vsim_command_lists = self.get_simulate_command_lists()
        util.write_shell_command_file(
            dirpath=self.args['work-dir'],
            filename='all.sh',
            command_lists = [['./pre_compile_dep_shell_commands.sh']] + vsim_command_lists
        )

        util.write_eda_config_and_args(dirpath=self.args['work-dir'], command_obj_ref=self)

    def compile(self):
        if self.args['stop-before-compile']:
            # don't run anything, save everyting we've already run in _prep_compile()
            return
        if self.args['stop-after-compile']:
            vsim_command_lists = self.get_compile_command_lists()
            self.run_commands_check_logs(vsim_command_lists, log_filename='sim.log')

    def elaborate(self):
        if self.args['stop-before-compile']:
            return
        if self.args['stop-after-compile']:
            return
        if self.args['stop-after-elaborate']:
        # only run this if we stop after elaborate (simulate run it all)
            vsim_command_lists = self.get_elaborate_command_lists()
            self.run_commands_check_logs(vsim_command_lists, log_filename='sim.log')

    def simulate(self):
        if self.args['stop-before-compile'] or self.args['stop-after-compile'] or \
           self.args['stop-after-elaborate']:
            # don't run this if we're stopping before/after compile/elab
            return
        vsim_command_lists = self.get_simulate_command_lists()
        self.run_commands_check_logs(vsim_command_lists, log_filename='sim.log')

    def get_compile_command_lists(self, **kwargs) -> list:
        # This will also set up a compile.
        vsim_command_list = [
            self.sim_exe,
            '' if self.args['gui'] else '-c',
            '-do', 'vsim_vlogonly.do', '-logfile', 'sim.log',
        ]
        return [vsim_command_list]

    def get_elaborate_command_lists(self, **kwargs) -> list:
        # This will also set up a compile, for vlog + vsim (0 time)
        vsim_command_list = [
            self.sim_exe,
            '' if self.args['gui'] else '-c',
            '-do', 'vsim_lintonly.do', '-logfile', 'sim.log',
        ]
        return [vsim_command_list]

    def get_simulate_command_lists(self, **kwargs) -> list:
    # This will also set up a compile, for vlog + vsim (with run -a)
        vsim_command_list = [
            self.sim_exe,
            '' if self.args['gui'] else '-c',
            '-do', 'vsim.do', '-logfile', 'sim.log',
        ]
        return [vsim_command_list]

    def get_post_simulate_command_lists(self, **kwargs) -> list:
        return []

    def write_vlog_dot_f(self, filename='vlog.f'):
        '''Returns none, creates filename (str) for a vlog.f'''
        vlog_dot_f_lines = []

        # Add compile args from config.tool.modelsim_ase:
        vlog_dot_f_lines += self.tool_config.get(
            'compile-args',
            '-sv -svinputport=net -lint').split()
        # Add waivers from config.tool.modelsim_ase:
        for waiver in self.tool_config.get(
                'compile-waivers',
                [ #defaults:
                    '2275', # 2275 - Existing package 'foo_pkg' will be overwritten.
                ]):
            vlog_dot_f_lines += ['-suppress', str(waiver)]

        vlog_dot_f_fname = filename
        vlog_dot_f_fpath = os.path.join(self.args['work-dir'], vlog_dot_f_fname)

        for value in self.incdirs:
            vlog_dot_f_lines += [ f"+incdir+{value}" ]

        for k,v in self.defines.items():
            if v is None:
                vlog_dot_f_lines += [ f'+define+{k}' ]
            else:

                # if the value v is a double-quoted string, such as v='"hi"', the
                # entire +define+NAME="hi" needs to wrapped in double quotes with the
                # value v double-quotes escaped: "+define+NAME=\"hi\""
                if isinstance(v, str) and v.startswith('"') and v.endswith('"'):
                    str_v = v.replace('"', '\\"')
                    vlog_dot_f_lines += [ f'"+define+{k}={str_v}"' ]
                else:
                    # Generally we should only support int and str python types passed as
                    # +define+{k}={v}, but also for SystemVerilog plusargs
                    vlog_dot_f_lines += [ f'+define+{k}={sanitize_defines_for_sh(v)}' ]


        vlog_dot_f_lines += self.args['compile-args']

        vlog_dot_f_lines += [
            '-source',
            ] + list(self.files_sv) + list(self.files_v)

        if len(self.files_sv) + len(self.files_v) == 0:
            self.error(f'{self.target=} {self.files_sv=} and {self.files_v=} are empty,',
                       'cannot create a valid vlog.f')

        with open(vlog_dot_f_fpath, 'w', encoding='utf-8') as f:
            f.writelines(line + "\n" for line in vlog_dot_f_lines)

    def write_vsim_dot_do(self, dot_do_to_write: list):
        '''Writes files(s) based on dot_do_to_write(list of str)

        list arg values can be empty (all) or have items 'all', 'sim', 'lint', 'vlog'.'''

        vsim_dot_do_fpath = os.path.join(self.args['work-dir'], 'vsim.do')
        vsim_lintonly_dot_do_fpath = os.path.join(self.args['work-dir'], 'vsim_lintonly.do')
        vsim_vlogonly_dot_do_fpath = os.path.join(self.args['work-dir'], 'vsim_vlogonly.do')

        sim_plusargs_str = self._get_sim_plusargs_str()
        vsim_suppress_list_str = self._get_vsim_suppress_list_str()

        voptargs_str = ""
        if self.args['gui'] or self.args['waves']:
            voptargs_str = self.tool_config.get('simulate-waves-args', '+acc')

        # TODO(drew): support self.args['sim_libary', 'elab-args', sim-args'] (3 lists)
        # to add to vsim_one_liner.

        vsim_one_liner = "vsim -onfinish stop" \
            + f" -sv_seed {self.args['seed']} {sim_plusargs_str} {vsim_suppress_list_str}" \
            + f" {voptargs_str} work.{self.args['top']}"

        vsim_one_liner = vsim_one_liner.replace('\n', ' ') # needs to be a one-liner

        vsim_vlogonly_dot_do_lines = [
            "if {[file exists work]} { vdel -all work; }",
            "vlib work;",
            "if {[catch {vlog -f vlog.f} result]} {",
            "    echo \"Caught $result \";",
            "    if {[batch_mode]} {",
            "        quit -f -code 20;",
            "    }",
            "}",
            "if {[batch_mode]} {",
            "    quit -f -code 0;",
            "}",
        ]

        vsim_lintonly_dot_do_lines = [
            "if {[file exists work]} { vdel -all work; }",
            "vlib work;",
            "quietly set qc 30;",
            "if {[catch {vlog -f vlog.f} result]} {",
            "    echo \"Caught $result \";",
            "    if {[batch_mode]} {",
            "        quit -f -code 20;",
            "    }",
            "}",
            "if {[catch { " + vsim_one_liner + " } result] } {",
            "    echo \"Caught $result\";",
            "    if {[batch_mode]} {",
            "        quit -f -code 19;",
            "    }",
            "}",
            "set TestStatus [coverage attribute -name SEED -name TESTSTATUS];",
            "if {[regexp \"TESTSTATUS += 0\" $TestStatus]} {",
            "    quietly set qc 0;",
            "} elseif {[regexp \"TESTSTATUS += 1\" $TestStatus]} {",
            "    quietly set qc 0;",
            "} else {",
            "    quietly set qc 2;",
            "}",
            "if {[batch_mode]} {",
            "    quit -f -code $qc;",
            "}",
        ]

        vsim_dot_do_lines = [
            "if {[file exists work]} { vdel -all work; }",
            "vlib work;",
            "quietly set qc 30;",
            "if {[catch {vlog -f vlog.f} result]} {",
            "    echo \"Caught $result \";",
            "    if {[batch_mode]} {",
            "        quit -f -code 20;",
            "    }",
            "}",
            "if {[catch { " + vsim_one_liner + " } result] } {",
            "    echo \"Caught $result\";",
            "    if {[batch_mode]} {",
            "        quit -f -code 19;",
            "    }",
            "}",
            "onbreak { resume; };",
            "catch {log -r *};",
            "run -a;",
            "set TestStatus [coverage attribute -name SEED -name TESTSTATUS];",
            "if {[regexp \"TESTSTATUS += 0\" $TestStatus]} {",
            "    quietly set qc 0;",
            "} elseif {[regexp \"TESTSTATUS += 1\" $TestStatus]} {",
            "    quietly set qc 0;",
            "} else {",
            "    quietly set qc 2;",
            "}",
            "if {[batch_mode]} {",
            "    quit -f -code $qc;",
            "}",
        ]

        write_all = len(dot_do_to_write) == 0 or 'all' in dot_do_to_write
        if write_all or 'sim' in dot_do_to_write:
            with open(vsim_dot_do_fpath, 'w', encoding='utf-8') as f:
                f.writelines(line + "\n" for line in vsim_dot_do_lines)

        if write_all or 'lint' in dot_do_to_write:
            with open(vsim_lintonly_dot_do_fpath, 'w', encoding='utf-8') as f:
                f.writelines(line + "\n" for line in vsim_lintonly_dot_do_lines)

        if write_all or 'vlog' in dot_do_to_write:
            with open(vsim_vlogonly_dot_do_fpath, 'w', encoding='utf-8') as f:
                f.writelines(line + "\n" for line in vsim_vlogonly_dot_do_lines)


    def _get_sim_plusargs_str(self) -> str:
        sim_plusargs = []

        assert isinstance(self.args["sim-plusargs"], list), \
            f'{self.target=} {type(self.args["sim-plusargs"])=} but must be list'

        for x in self.args['sim-plusargs']:
            # For vsim we need to add a +key=value if the + is missing
            if x[0] != '+':
                x = f'+{x}'
            sim_plusargs.append(x)

        return ' '.join(sim_plusargs)


    def _get_vsim_suppress_list_str(self) -> str:
        vsim_suppress_list = []
        # Add waivers from config.tool.modelsim_ase:
        for waiver in self.tool_config.get(
                'simulate-waivers', [
                    #defaults:
                    '3009', # 3009: [TSCALE] - Module 'foo' does not have a timeunit/timeprecision
                            #       specification in effect, but other modules do.
                ]):
            vsim_suppress_list += ['-suppress', str(waiver)]

        return ' '.join(vsim_suppress_list)


class CommandElabModelsimAse(CommandSimModelsimAse):
    '''CommandElabModelsimAse is a command handler for: eda elab --tool=modelsim_ase'''

    def __init__(self, config:dict):
        super().__init__(config)
        self.args['stop-after-elaborate'] = True



'''opencos.commands.waves - command handler for: eda waves ...

Note this command is handled differently than others (such as CommandSim),
it is generally run as simply

    > eda waves

and attempts to auto-find the newest waveform in your work directories. As
a result, no Tool is bound to CommandWaves (there is no CommandWavesVivado
handler).
'''

# Note - similar code waiver, tricky to eliminate it with inheritance when
# calling reusable methods.
# pylint: disable=R0801

import os
import shutil

from opencos import util
from opencos.eda_base import CommandDesign


class CommandWaves(CommandDesign):
    '''command handler for: eda waves'''

    command_name = 'waves'

    SUPPORTED_WAVES_EXT = [
        '.wdb', '.vcd', '.wlf', '.fst'
    ]

    VSIM_TOOLS = set([
        'questa',
        'modelsim_ase',
    ])

    VSIM_VCD_TOOLS = set([
        'questa',
    ])

    def __init__(self, config: dict):
        CommandDesign.__init__(self, config=config)
        self.args.update({
            'test-mode': False,
        })
        self.args_help.update({
            'test-mode': 'Do not run the command to open the located wave file, instead print' \
            + ' to stdout',
        })


    def get_wave_files_in_dirs(self, wave_dirs: list, quiet: bool = False) -> list:
        '''Returns list of all wave files give wave_dirs (list)'''

        def info(*text):
            if not quiet:
                util.info(*text)

        all_files = []
        for d in wave_dirs:
            info(f"Looking for wavedumps below: {d}")
            for root, _, files in os.walk(d):
                for f in files:
                    for e in self.SUPPORTED_WAVES_EXT:
                        if f.endswith(e):
                            info(f"Found wave file: {os.path.join(root, f)}")
                            all_files.append(os.path.join(root, f))
        return all_files

    def process_tokens( # pylint: disable=too-many-branches,too-many-statements
            self, tokens: list, process_all: bool = True,
            pwd: str = os.getcwd()
    ) -> list:

        wave_file = None
        wave_dirs = []
        tokens = CommandDesign.process_tokens(self, tokens=tokens, process_all=False, pwd=pwd)

        if self.args['test-mode']:
            self.exec = self._test_mode_exec

        while tokens:
            if os.path.isfile(tokens[0]):
                if wave_file is not None:
                    self.error(f"Was already given {wave_file=}, not sure what",
                               f"to do with: {tokens[0]}")
                wave_file = os.path.abspath(tokens[0])
                tokens.pop(0)
                continue
            if os.path.isdir(tokens[0]):
                if wave_file is not None:
                    self.error(f"Was already given {wave_file=}, not sure what",
                               f"to do with {tokens[0]}")
                wave_dirs.append(tokens[0])
            self.error(f"Didn't understand command arg/token: '{tokens[0]}'",
                       "in CommandWaves")

        if not wave_file:
            util.info("need to look for wave file")
            # we weren't given a wave file, so we will look for one!
            if not wave_dirs and os.path.isdir(self.args['eda-dir']):
                wave_dirs.append(self.args['eda-dir'])
            if not wave_dirs:
                wave_dirs.append('.')
            all_files = self.get_wave_files_in_dirs(wave_dirs)
            if len(all_files) > 1:
                all_files.sort(key=os.path.getmtime)
                util.info(f"Choosing: {all_files[-1]} (newest)")
            if all_files:
                wave_file = all_files[-1]
            else:
                self.error(f"Couldn't find any wave files below: {','.join(wave_dirs)}")

        wave_file = os.path.abspath(wave_file)
        util.info(f"decided on opening: {wave_file}")

        # TODO(drew): this feels a little customized per-tool, perhaps there's a better
        # way to abstract this configuration for adding other waveform viewers.
        # For example for each command we also have to check shutil.which, because normal Tool
        # classs should work even w/out PATH, but these don't use Tool classes.
        if wave_file.endswith('.wdb'):
            if 'vivado' in self.config['tools_loaded'] and shutil.which('vivado'):
                tcl_name = wave_file + '.waves.tcl'
                with open( tcl_name, 'w', encoding='utf-8') as fo :
                    print( 'current_fileset', file=fo)
                    print( f'open_wave_database {wave_file}', file=fo)
                command_list = [ 'vivado', '-source', tcl_name]
                self.exec(os.path.dirname(wave_file), command_list)
            else:
                self.error(f"Don't know how to open {wave_file} without Vivado in PATH")
        elif wave_file.endswith('.wlf'):
            if self._vsim_available():
                command_list = ['vsim', wave_file]
                self.exec(os.path.dirname(wave_file), command_list)
            else:
                self.error(f"Don't know how to open {wave_file} without one of",
                           f"{self.VSIM_TOOLS} in PATH")
        elif wave_file.endswith('.fst'):
            if 'gtkwave' in self.config['tools_loaded'] and shutil.which('gtkwave'):
                command_list = ['gtkwave', wave_file]
                self.exec(os.path.dirname(wave_file), command_list)
            else:
                self.error(f"Don't know how to open {wave_file} without GtkWave in PATH")
        elif wave_file.endswith('.vcd'):
            if 'gtkwave' in self.config['tools_loaded'] and shutil.which('gktwave'):
                command_list = ['gtkwave', wave_file]
                self.exec(os.path.dirname(wave_file), command_list)
            elif self._vsim_available(from_tools=self.VSIM_VCD_TOOLS):
                # TODO(drew): untested, may not work, may need to use fst2vcd converter first
                # (from gtkwave install)
                command_list = ['vsim', wave_file]
                self.exec(os.path.dirname(wave_file), command_list)
            else:
                self.error(f"Don't know how to open {wave_file} without Vivado,",
                           f"gtkwave, or {self.VSIM_VCD_TOOLS} in PATH")

        return tokens

    def _vsim_available( # pylint: disable=dangerous-default-value
            self, from_tools: list = VSIM_TOOLS
    ) -> bool:
        '''Returns True if 'vsim' is available (Questa or Modelsim)'''
        return bool(shutil.which('vsim')) and \
            any(x in self.config['tools_loaded'] for x in from_tools)


    def _test_mode_exec( # pylint: disable=unused-argument
            self, work_dir: str,
            command_list: list,
            **kwargs
    ) -> None:
        '''Override for Command.exec if arg --test-mode was set, does not run

        the command_list, instead prints to stdout'''

        util.info(f'waves.py: test_mode exec stdout: {" ".join(command_list)};',
                  f'   ({work_dir=}')



''' opencos.tools.invio - class for ToolInvio to determine if Invio is present and what version'''

# pylint: disable=R0801 # (duplicate code in derived classes, such as if-condition return.)

import importlib.util

from opencos import util
from opencos.tools import invio_helpers
from opencos.eda_base import Tool
from opencos.commands import CommandElab



class ToolInvio(Tool):
    '''Invio w/out Yosys, used for elab in SIMULATIION (not the same as ToolInvioYosys)'''

    _TOOL = 'invio'
    _URL = 'https://www.verific.com/products/invio/'

    def get_versions(self) -> str:
        if self._VERSION:
            return self._VERSION

        # We also have to make sure we can import invio within python.
        spec = importlib.util.find_spec('invio')
        if not spec:
            self.error('"invio" package not in python env')


        return super().get_versions()

    def set_tool_defines(self):
        super().set_tool_defines()
        self.defines.update({
            'OC_TOOL_INVIO': None,
        })

class CommandElabInvio(CommandElab, ToolInvio):
    '''Command handler for: eda elab --tool=invio'''

    def __init__(self, config:dict):
        CommandElab.__init__(self, config)
        ToolInvio.__init__(self, config=self.config)
        self.args.update({
            'invio-blackbox': [],                # list of modules that invio/verific will blackbox.
        })

        self.invio_command_lists = []

    # Note that we follow parent class CommandSim's do_it() flow, that way --export args
    # are handled.
    def prepare_compile(self) -> None:
        ''' prepare_compile() - following parent Commandsim's run() flow'''
        self.set_tool_defines()
        self.write_eda_config_and_args()

        self.invio_command_lists = self.get_compile_command_lists()
        self.write_invio_sh()

    def compile(self) -> None:
        pass

    def elaborate(self) -> None:
        ''' elaborate() - following parent Commandsim's run() flow, runs invio_command_lists'''
        if self.args['stop-before-compile'] or \
           self.args['stop-after-compile']:
            return
        # Finally, run the command(s) if we made it this far: python run_invio.py:
        self.run_commands_check_logs(self.invio_command_lists)

    def get_compile_command_lists(self, **kwargs) -> list:
        '''Returns list of util.ShellCommandList, for slang we'll run this in elaborate()'''
        invio_blackbox_list = self.args.get('invio-blackbox', [])
        invio_dict = invio_helpers.get_invio_command_dict(
            self, blackbox_list=invio_blackbox_list, sim_elab=True
        )
        return invio_dict['command_lists']

    def write_invio_sh(self) -> None:
        '''Returns None, writes out run_invio.sh for reproducing outside of eda framework.'''
        util.write_shell_command_file(dirpath=self.args['work-dir'], filename='run_invio.sh',
                                      command_lists=self.invio_command_lists, line_breaks=True)

    def get_elaborate_command_lists(self, **kwargs) -> list:
        '''We only use 'compile' commands for invio elab, do not use elaborate commands'''
        return []




import os
import sys

from opencos import eda, eda_config, util

# Used by pytest, so we can skip tests if tools aren't present.

def get_config_and_tools_loaded(quiet=False, args=[]):
    # We have to figure out what tools are avaiable w/out calling eda.main,
    # so we can get some of these using eda_config.get_eda_config()
    config, _ = eda_config.get_eda_config(args=args, quiet=quiet)
    config = eda.init_config(config=config, quiet=quiet)
    tools_loaded = config.get('tools_loaded', set()).copy()
    return config, tools_loaded

def get_all_handler_commands(config=None, tools_loaded=None) -> dict:
    all_handler_commands = dict()

    if config is None or tools_loaded is None:
        config, tools_loaded = get_config_and_tools_loaded()

    assert type(config) is dict
    assert type(tools_loaded) is set

    # Let's re-walk auto_tools_order to get this ordered per eda command:
    for tool, table in config.get('auto_tools_order', [{}])[0].items():
        if tool not in tools_loaded:
            continue

        if table.get('disable-tools-multi', False):
            # Flagged as do-not-add when running eda command: tools-multi
            util.debug(f'eda_tool_helper.py -- skipping {tool=} it is set with flag',
                       'disable-tools-multi in config')
            continue

        for command, handler in table.get('handlers', {}).items():
            if command not in all_handler_commands:
                # create ordered list from config.
                all_handler_commands[command] = list([tool])
            else:
                all_handler_commands[command].append(tool)

    return all_handler_commands
